package coco.codegen.prolog;

import java.util.*;
import java.io.*;

import coco.codegen.*;
import coco.utils.*;
import coco.parser.*;
import coco.parser.ast.*;
import coco.parser.symtab.*;




public class PrologGenerator extends AbstractASTVisitor implements AbstractCodeGenerator, CoCoParserConstants {

    // This stores the generated text
    StringWriter result;
    PrintWriter writer;

    // State for the visitor methods
    Stack context;

    /**
     * Standard constructor initializes the generator.
     */
    public PrologGenerator() {
	init();
    }

    // Methods from AbstractCodeGenerator

    /** 
     * This method initializes the code generator.
     */
    public void init() {
       	result = new StringWriter();
	writer = new PrintWriter(result);

	context = new Stack();

	// start work
	generateProlog();
    }

    /**
     * This method retrieves the result of the CodeGenerator
     */
    public String getResult() {
	writer.flush();
	return(result.toString());
    }

    /**
     * This method starts the prolog generation process.
     * All compilation units from the symbol table are used.
     * After gerenation has been completed, the result can be fetched using the getResult method.
     */
    private void generateProlog() {

	plComment("This file is automatically generated at " + new Date() + ".");
	

	// built in types

	plComment("List of built-in types.");
	plDataType(removeQuotes(tokenImage[BOOL]));
	plBuiltInType(removeQuotes(tokenImage[BOOL]));
	plDataType(removeQuotes(tokenImage[BYTE]));
	plBuiltInType(removeQuotes(tokenImage[BYTE]));
	plDataType(removeQuotes(tokenImage[CHAR]));
	plBuiltInType(removeQuotes(tokenImage[CHAR]));
	plDataType(removeQuotes(tokenImage[DOUBLE]));
	plBuiltInType(removeQuotes(tokenImage[DOUBLE]));
	plDataType(removeQuotes(tokenImage[FLOAT]));
	plBuiltInType(removeQuotes(tokenImage[FLOAT]));
	plDataType(removeQuotes(tokenImage[INT]));
	plBuiltInType(removeQuotes(tokenImage[INT]));
	plDataType(removeQuotes(tokenImage[LONG]));
	plBuiltInType(removeQuotes(tokenImage[LONG]));
	plDataType(removeQuotes(tokenImage[SHORT]));
	plBuiltInType(removeQuotes(tokenImage[SHORT]));
	plDataType(removeQuotes(tokenImage[VOID]));
	plBuiltInType(removeQuotes(tokenImage[VOID]));

	plComment("List of components.");
	// all components
	for (Enumeration enum = SymtabService.getComponents(); enum.hasMoreElements(); ) {
	    ASTComponent c = (ASTComponent)enum.nextElement();
	    Assert.require(c);

	    plComment("Component " + c.getName());
	    c.accept(this, null);
	}

	plComment("List of interfaces.");
	// all interfaces
	for (Enumeration enum = SymtabService.getInterfaces(); enum.hasMoreElements(); ) {
	    ASTInterface c = (ASTInterface)enum.nextElement();
	    Assert.require(c);

	    plComment("Interface " + c.getID().getID());
	    c.accept(this, null);
	}

	plComment("List of property sets.");
	// all property sets
	for (Enumeration enum = SymtabService.getPropertySets(); enum.hasMoreElements(); ) {
	    ASTPropertySet c = (ASTPropertySet)enum.nextElement();
	    Assert.require(c);

	    plComment("Property set " + c.getId().getID());
	    c.accept(this, null);
	}

	plComment("List of data types.");
	// all data types
	for (Enumeration enum = SymtabService.getDataTypes(); enum.hasMoreElements(); ) {
	    ASTDataType  c = (ASTDataType)enum.nextElement();
	    Assert.require(c);

	    plComment("Datatype " + c.getDataTypeIdentifier().getID());
	    c.accept(this, null);
	}
    }

    // Methods from AbstractCodeGenerator

    /** 
     * This method tells, if this visitor wants to be pre-visited.
     */
    public boolean preVisit() {return(true);}

    /** 
     * This method tells, if this visitor wants to be post-visited.
     */
    public boolean postVisit() {return(true);}


    // Visitor methods
    
    // ComponentList
    public Object visitPre(ASTComponentList node) {return(null);}
    public Object visitIntraList(ASTComponentList node) {return(null);}
    public Object visitPost(ASTComponentList node) {return(null);}

    // Component
    public Object visitPre(ASTComponent node) {
	context.push(node);

	plComponent(node.getName());

	if (node.isActive()) {plActiveComponent(node.getName());}

	return(null);
    }
    public Object visitPost(ASTComponent node) {
	context.pop();
	return(null);
    }

    // implements Interface
    public Object visitPre(ASTInterfaceIdentifierList node) {
	context.push(new Vector());
	return(null);
    }
    public Object visitIntraList(ASTInterfaceIdentifierList node) {return(null);}
    public Object visitPost(ASTInterfaceIdentifierList node) {
	Vector vec = (Vector)context.pop();
	String[] names = (String[])vec.toArray(new String[0]);

	if (context.peek() instanceof ASTComponent) {
	    String compName = ((ASTComponent)context.peek()).getName();
	    plImplements(compName, names);
	} else {
	}
	return(null);
    }

    // has PropertySet
    public Object visitPre(ASTPropertySetReferenceList node) {
	context.push(new Vector());
	return(null);
    }
    public Object visitIntraList(ASTPropertySetReferenceList node) {return(null);}
    public Object visitPost(ASTPropertySetReferenceList node) {
	Vector vec = (Vector)context.pop();
	String[] names = (String[])vec.toArray(new String[0]);

	if (context.peek() instanceof ASTComponent) {
	    String compName = ((ASTComponent)context.peek()).getName();
	    plHas(compName, names);

	} else if (context.peek() instanceof ASTPort) {
	    String compName = ((ASTPort)context.peek()).getPortId().getID();
	    plHas(compName, names);

	} else {
	}
	return(null);
    }

    // ComponentBody
    public Object visitPre(ASTComponentBody node) {
	return(null);
    }
    public Object visitPost(ASTComponentBody node) {
	return(null);
    }

    // Interface
    public Object visitPre(ASTInterface node) {
	plInterface(node.getID().getID());

	// store current node
	context.push(node);
	return(null);
    }

    public Object visitPost(ASTInterface node) {
	// remove current node
	context.pop();
	return(null);
    }

    // Interface body
    public Object visitPre(ASTInterfaceBody node) {return(null);}
    public Object visitPost(ASTInterfaceBody node) {return(null);}


    // Datatype
    public Object visitPre(ASTDataType node) {
	String dName = node.getDataTypeIdentifier().getID();
	plDataType(dName);

	if (node.getBasetypeIdentifier() != null) {
	    String bName = node.getBasetypeIdentifier().getID();
	    plBaseType(dName, bName);
	}

	context.push(node);
	return(null);
    }
    public Object visitPost(ASTDataType node) {
	context.pop();
	return(null);
    }

    // Field
    public Object visitPre(ASTField node) {
	String fName = node.getFieldId().getID();
	String fType = node.getTypeId().getTypeID().getID();

	ASTDataTypeIdentifier dTypeId = node.getTypeId();
	String[] fDims = {};
	if (dTypeId.isArrayType() && (dTypeId.getDimensions() != null)) {
	    fDims = new String[dTypeId.getDimensions().size()];
	    for (int i = 0; i < dTypeId.getDimensions().size(); i++) {fDims[i] = ((Integer)dTypeId.getDimensions().elementAt(i)).toString();}
	}

	if (context.peek() instanceof ASTDataType) {
	    String compName = ((ASTDataType)context.peek()).getDataTypeIdentifier().getID();
	    plField(compName, fType, fDims, fName);
	} else {
	}
	
	return(null);
    }
    public Object visitPost(ASTField node) {return(null);}




    public Object visitPre(ASTComponentTypeIdentifier node) {return(null);}
    public Object visitPost(ASTComponentTypeIdentifier node) {return(null);}

    public Object visitPre(ASTRoleIdentifier node) {return(null);}
    public Object visitPost(ASTRoleIdentifier node) {return(null);}

    public Object visitPre(ASTInterfaceIdentifier node) {return(visitPre((ASTIdentifier)node));}
    public Object visitPost(ASTInterfaceIdentifier node) {return(visitPost((ASTIdentifier)node));}




    public Object visitPre(ASTPortPropertySet node) {
	plPortPropertySet(node.getId().getID());

	context.push(node);
	return(null);
    }
    public Object visitPost(ASTPortPropertySet node) {
	context.pop();
	return(null);
    }

    public Object visitPre(ASTComponentPropertySet node) {
	plComponentPropertySet(node.getId().getID());

	context.push(node);
	return(null);
    }
    public Object visitPost(ASTComponentPropertySet node) {
	context.pop();
	return(null);
    }

    public Object visitPre(ASTConnectorPropertySet node) {return(null);}
    public Object visitPost(ASTConnectorPropertySet node) {return(null);}




    public Object visitPre(ASTHandler node) {return(null);}
    public Object visitPost(ASTHandler node) {return(null);}




    public Object visitPre(ASTNumericLiteral node) {return(null);}
    public Object visitPost(ASTNumericLiteral node) {return(null);}

    public Object visitPre(ASTInvocationTypeReference node) {return(null);}
    public Object visitPost(ASTInvocationTypeReference node) {return(null);}

    public Object visitPre(ASTHandlerTypeReference node) {return(null);}
    public Object visitPost(ASTHandlerTypeReference node) {return(null);}



    // See ASTDesignator
    public Object visitPre(ASTPortAccess node) {return(null);}
    public Object visitPost(ASTPortAccess node) {

	// this is done by ASTDesignator
	String str = (String)context.pop();

	if (context.peek() instanceof Vector) {
	    // we are in a list 
	    Vector v = (Vector)context.peek();
	    v.add(str);
	} else {
	    context.push(str);
	}

	return(null);
    }




    public Object visitPre(ASTPortTypeList node) {return(null);}

    public Object visitIntraList(ASTPortTypeList node) {return(null);}


    public Object visitPost(ASTPortTypeList node) {return(null);}







    public Object visitPre(ASTIdentifierList node) {

	// hack 
	// Method visitXXX(ASTPropertySetReferenceList node) is not defined in AbstractASTVisitor
	if (node instanceof ASTPropertySetReferenceList) {
	    visitPre((ASTPropertySetReferenceList)node);
	} else {
       	    context.push(new Vector());
	}
	return(null);
    }
    public Object visitIntraList(ASTIdentifierList node) {	
	// hack 
	// Method visitXXX(ASTPropertySetReferenceList node) is not defined in AbstractASTVisitor
	if (node instanceof ASTPropertySetReferenceList) {
	    visitIntraList((ASTPropertySetReferenceList)node);
	} else {
	}
	return(null);
    }
    public Object visitPost(ASTIdentifierList node) {
	// hack 
	// Method visitXXX(ASTPropertySetReferenceList node) is not defined in AbstractASTVisitor
	if (node instanceof ASTPropertySetReferenceList) {
	    visitPost((ASTPropertySetReferenceList)node);
	} else {
	
	    Vector vec = (Vector)context.pop();
	    String[] names = (String[])vec.toArray(new String[0]);
	    
	    if (context.peek() instanceof ASTInterface) {
		String compName = ((ASTInterface)context.peek()).getID().getID();
		plExtends(compName, names);
	    } else {
	    }
	}
	return(null);
    }


    public Object visitPre(ASTPropertyContainer node) {return(null);}
    public Object visitPost(ASTPropertyContainer node) {return(null);}


    public Object visitPre(ASTPortModifiers node) {return(null);}
    public Object visitPost(ASTPortModifiers node) {return(null);}

    public Object visitPre(ASTIdentifier node) {
	if (context.peek() instanceof Vector) {
	    // we are in a list
	    ((Vector)context.peek()).add(node.getID());
	} else {
	}
	return(null);
    }
    public Object visitPost(ASTIdentifier node) {return(null);}


    public Object visitPre(ASTPropertyIdentifier node) {return(null);}
    public Object visitPost(ASTPropertyIdentifier node) {return(null);}

    public Object visitPre(ASTCharLiteral node) {return(null);}
    public Object visitPost(ASTCharLiteral node) {return(null);}


    public Object visitPre(ASTPortTypeReferenceList node) {return(null);}

    public Object visitIntraList(ASTPortTypeReferenceList node) {return(null);}



    public Object visitPost(ASTPortTypeReferenceList node) {return(null);}


    public Object visitPre(ASTInvocationSignature node) {return(null);}
    public Object visitPost(ASTInvocationSignature node) {return(null);}


    public Object visitPre(ASTUserTypeIdentifier node) {return(null);}
    public Object visitPost(ASTUserTypeIdentifier node) {return(null);}

    public Object visitPre(ASTBaseTypeIdentifier node) {return(null);}
    public Object visitPost(ASTBaseTypeIdentifier node) {return(null);}

    public Object visitPre(ASTBaseTypeReference node) {return(null);}
    public Object visitPost(ASTBaseTypeReference node) {return(null);}

    public Object visitPre(ASTInvocationParameterList node) {return(null);}

    public Object visitIntraList(ASTInvocationParameterList node) {return(null);}


    public Object visitPost(ASTInvocationParameterList node) {return(null);}




    public Object visitPre(ASTBoolLiteral node) {return(null);}
    public Object visitPost(ASTBoolLiteral node) {return(null);}



    public Object visitPre(ASTConnector node) {
	// store context
	context.push(node);
	return(null);
    }

    public Object visitPost(ASTConnector node) {

	// Hack:
	// see ASTPortAccessList
	
	// pop the second copy
	context.pop();

	// get the port access list 
	Vector v = (Vector)context.pop();
	String cName = node.getConnectorID().getID();

	// remove context
	context.pop();

	if (context.peek() instanceof ASTComponent) {
	    String name = ((ASTComponent)context.peek()).getName();
	    plConnector(name, cName, (String[])v.toArray(new String[0]));

	} else if (context.peek() instanceof ASTInterface) {
	    String name = ((ASTInterface)context.peek()).getID().getID();
	    plConnector(name, cName, (String[])v.toArray(new String[0]));

	} else {
	}



	return(null);
    }

    public Object visitPre(ASTConnectorBody node) {
	writer.write(").\n");
	return(null);
    }
    public Object visitPost(ASTConnectorBody node) {return(null);}



    public Object visitPre(ASTConstExpression node) {return(null);}
    public Object visitPost(ASTConstExpression node) {return(null);}


    public Object visitPre(ASTProperty node) {
	/*
	String pName = node.getPropertyId().getID();
	String pMod = "mandatory";
	if (node.getVisibility() == ASTProperty.OPTIONAL) {pMod = "optional";}
	String pVal = "";
	if (node.getInitExpr() != null) {pVal = node.getInitExpr().toString();}

	if (context.peek() instanceof ASTComponent) {
	    String name = ((ASTComponent)context.peek()).getName();
	    plProperty(name, pMod, pName, pVal);

	} else if (context.peek() instanceof ASTInterface)  {
	    String name = ((ASTInterface)context.peek()).getID().getID();
	    plProperty(name, pMod, pName, pVal);
	
	} else if (context.peek() instanceof ASTConnector)  {
	    String name = ((ASTConnector)context.peek()).getConnectorID().getID();
	    plProperty(name, pMod, pName, pVal);

	} else if (context.peek() instanceof ASTPropertySet)  {
	    String name = ((ASTPropertySet)context.peek()).getId().getID();
	    plProperty(name, pMod, pName, pVal);
	}
	*/
	return(null);
    }
    public Object visitPost(ASTProperty node) {

	// Done by ASTDesignator
	String pName = (String)context.pop();
	String pMod = "mandatory";
	if (node.getVisibility() == ASTProperty.OPTIONAL) {pMod = "optional";}
	String pVal = "";
	if (node.getInitExpr() != null) {pVal = node.getInitExpr().toString();}

	if (context.peek() instanceof ASTComponent) {
	    String name = ((ASTComponent)context.peek()).getName();
	    plProperty(name, pMod, pName, pVal);

	} else if (context.peek() instanceof ASTInterface)  {
	    String name = ((ASTInterface)context.peek()).getID().getID();
	    plProperty(name, pMod, pName, pVal);
	
	} else if (context.peek() instanceof ASTConnector)  {
	    String name = ((ASTConnector)context.peek()).getConnectorID().getID();
	    plProperty(name, pMod, pName, pVal);

	} else if (context.peek() instanceof ASTPropertySet)  {
	    String name = ((ASTPropertySet)context.peek()).getId().getID();
	    plProperty(name, pMod, pName, pVal);
	}
	return(null);
    }






    public Object visitPre(ASTPortBody node) {return(null);}
    public Object visitPost(ASTPortBody node) {return(null);}

    public Object visitPre(ASTFloatLiteral node) {return(null);}
    public Object visitPost(ASTFloatLiteral node) {return(null);}


    public Object visitPre(ASTPortList node) {return(null);}
    public Object visitIntraList(ASTPortList node) {return(null);}
    public Object visitPost(ASTPortList node) {return(null);}


    // Designators are ASTIdenitfiers with dots in between
    public Object visitPre(ASTDesignator node) {
	context.push(new Vector());
	return(null);
    }
    public Object visitIntraList(ASTDesignator node) {return(null);}
    public Object visitPost(ASTDesignator node) {
	Vector v = (Vector)context.pop();
	
	StringBuffer str = new StringBuffer();
	for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
	    str.append((String)e.nextElement());
	    if (e.hasMoreElements()) {str.append(".");}
	}

	context.push(str.toString());

	return(null);
    }



    public Object visitPre(ASTConnectorList node) {return(null);}

    public Object visitIntraList(ASTConnectorList node) {return(null);}



    public Object visitPost(ASTConnectorList node) {return(null);}


    public Object visitPre(ASTPortAccessList node) {
	context.push(new Vector());
	return(null);
    }
    public Object visitIntraList(ASTPortAccessList node) {return(null);}
    public Object visitPost(ASTPortAccessList node) {

	// Hack:
	// To maintain the context and to provide the reult of the 
	// PortAccessList to the ASTConnector method, we must place a second copy of the connector 
	// pointer onto the stack
	// ASTConnector method has to remove both of these objcets

	Vector v = (Vector)context.pop();

	if (context.peek() instanceof ASTConnector) {
	    ASTConnector con = (ASTConnector)context.peek();
	    context.push(v);
	    context.push(con);
	}

	return(null);
    }



    public Object visitPre(ASTStringLiteral node) {return(null);}
    public Object visitPost(ASTStringLiteral node) {return(null);}


    public Object visitPre(ASTConstExpressionList node) {return(null);}
    public Object visitIntraList(ASTConstExpressionList node) {return(null);}



    public Object visitPost(ASTConstExpressionList node) {return(null);}


    public Object visitPre(ASTPortTypeReference node) {return(null);}
    public Object visitPost(ASTPortTypeReference node) {return(null);}


    public Object visitPre(ASTInvocationParameter node) {return(null);}
    public Object visitPost(ASTInvocationParameter node) {return(null);}


    public Object visitPre(ASTInstanceList node) {return(null);}
    public Object visitIntraList(ASTInstanceList node) {return(null);}
    public Object visitPost(ASTInstanceList node) {return(null);}

    public Object visitPre(ASTRoleList node) {return(null);}
    public Object visitIntraList(ASTRoleList node) {return(null);}
    public Object visitPost(ASTRoleList node) {return(null);}



    public Object visitPre(ASTRole node) {
	String rName = node.getVarId().getID();
	String rType = node.getTypeId().getID();

	if (context.peek() instanceof ASTInterface) {
	    String name = ((ASTInterface)context.peek()).getID().getID();
	    plRole(name, rType, rName);
	    
	} else {
	}

	return(null);
    }
    public Object visitPost(ASTRole node) {return(null);}



    public Object visitPre(ASTFieldList node) {return(null);}

    public Object visitIntraList(ASTFieldList node) {return(null);}


    public Object visitPost(ASTFieldList node) {return(null);}



    public Object visitPre(ASTInstance node) {

	String iName = node.getVarId().getID();
	String iType = node.getTypeId().getID();
	String iRole = "";
	if (node.getRoleId() != null) {iRole = node.getRoleId().getID();}
	
	if (context.peek() instanceof ASTComponent) {
	    String name = ((ASTComponent)context.peek()).getName();
	    plInstance(name, iType, iName, iRole);

	} else if (context.peek() instanceof ASTInterface)  {
	    String name = ((ASTInterface)context.peek()).getID().getID();
	    plInstance(name, iType, iName, iRole);

	}
	return(null);
    }
    public Object visitPost(ASTInstance node) {return(null);}




    public Object visitPre(ASTData node) {

	//get information about the port

	String pName = node.getPortId().getID();
	String pType = node.getDataType().getTypeID().getID();
	String pDir = node.getDirection();
	String[] pMods = {};
	if (node.getModifiers() != null) {pMods = new String[] {node.getModifiers().getFirstModifierString(), node.getModifiers().getSecondModifierString()};}

	ASTDataTypeIdentifier dTypeId = node.getDataType();
	String[] pDims = {};
	if (dTypeId.isArrayType() && (dTypeId.getDimensions() != null)) {
	    pDims = new String[dTypeId.getDimensions().size()];
	    for (int i = 0; i < dTypeId.getDimensions().size(); i++) {pDims[i] = ((Integer)dTypeId.getDimensions().elementAt(i)).toString();}
	}

	if (context.peek() instanceof ASTComponent) {
	    String name = ((ASTComponent)context.peek()).getName();
	    plPort(name, pMods, pDir, pType, pDims, pName);

	} else if (context.peek() instanceof ASTInterface)  {
	    String name = ((ASTInterface)context.peek()).getID().getID();
	    plPort(name, pMods, pDir, pType, pDims, pName);

	}

	context.push(node);
	
	return(null);
    }

    public Object visitPost(ASTData node) {
	context.pop();
	return(null);
    }





    public Object visitPre(ASTCompilationUnit node) {return(null);}
    public Object visitPost(ASTCompilationUnit node) {return(null);}


    public Object visitPre(ASTIntLiteral node) {return(null);}
    public Object visitPost(ASTIntLiteral node) {return(null);}




    public Object visitPre(ASTPort node) {return(null);}
    public Object visitPost(ASTPort node) {return(null);}


    public Object visitPre(ASTDataTypeReference node) {return(null);}
    public Object visitPost(ASTDataTypeReference node) {return(null);}



    public Object visitPre(ASTPortType node) {return(null);}
    public Object visitPost(ASTPortType node) {return(null);}



    public Object visitPre(ASTDataTypeList node) {return(null);}

    public Object visitIntraList(ASTDataTypeList node) {return(null);}



    public Object visitPost(ASTDataTypeList node) {return(null);}


    public Object visitPre(ASTPropertyList node) {return(null);}
    public Object visitIntraList(ASTPropertyList node) {return(null);}



    public Object visitPost(ASTPropertyList node) {return(null);}


    public Object visitPre(ASTGenericTypeIdentifierList node) {return(null);}



    public Object visitPost(ASTGenericTypeIdentifierList node) {return(null);}



    public Object visitPre(ASTDataTypeIdentifierList node) {return(null);}
    public Object visitIntraList(ASTDataTypeIdentifierList node) {return(null);}



    public Object visitPost(ASTDataTypeIdentifierList node) {return(null);}


    public Object visitPre(ASTEntititySpec node) {return(null);}
    public Object visitPost(ASTEntititySpec node) {return(null);}

    public Object visitPre(ASTGenericTypeParameter node) {return(null);}
    public Object visitPost(ASTGenericTypeParameter node) {return(null);}

    public Object visitPre(ASTMappingLiteral node) {return(null);}
    public Object visitIntraList(ASTMappingLiteral node) {return(null);}



    public Object visitPost(ASTMappingLiteral node) {return(null);}

    public Object visitPre(ASTMapping node) {return(null);}
    public Object visitPost(ASTMapping node) {return(null);}



    public Object visitPre(ASTExternalTypeName node) {return(null);}
    public Object visitPost(ASTExternalTypeName node) {return(null);}


    public Object visitPre(ASTRoot node) {return(null);}
    public Object visitPost(ASTRoot node) {return(null);}


    public Object visitPre(ASTGenericTypeParameterList node) {return(null);}

    public Object visitIntraList(ASTGenericTypeParameterList node) {return(null);}


    public Object visitPost(ASTGenericTypeParameterList node) {return(null);}



    // prolog generating methods

    public void plComment(String str) {
	writer.println("% " + str);
    }

    public void plComponent(String name) {
	writer.println("component('" + name + "').");
    }

    public void plActiveComponent(String name) {
	writer.println("activecomponent('" + name + "').");
    }

    public void plImplements(String compName, String[] intfNames) {
	writer.println("implements('" + compName + "', [" + makeStringList(intfNames) + "]).");
    }

    public void plHas(String compName, String[] propNames) {
	writer.println("has('" + compName + "', [" + makeStringList(propNames) + "]).");
    }

    public void plPort(String compName, String[] mods, String dir, String type, String[] dims, String name) {
	writer.println("port('" + compName + "', [" + makeStringList(mods) + "], '" + dir + "', '" + type + "', [" + makeStringList(dims) + "], '" + name + "').");
    }

    public void plProperty(String compName, String mod, String name, String wert) {
	writer.println("property('" + compName + "', '" + mod + "', '" + name + "', '" + wert + "').");
    }

    public void plInstance(String compName, String type, String name, String role) {
	writer.println("instance('" + compName + "', '" + type + "', '" + name + "', '" + role + "').");
    }

    public void plConnector(String compName, String name, String[] portNames) {
	writer.println("connector('" + compName + "', [" + makeStringList(portNames) + "]).");
    }
    
    public void plInterface(String name) {
	writer.println("abstractcomponent('" + name + "').");
    }

    public void plExtends(String compName, String[] intfNames) {
	writer.println("extends('" + compName + "', [" + makeStringList(intfNames) + "]).");
    }
    
    public void plRole(String compName, String type, String name) {
	writer.println("role('" + compName + "', '" + type + "', '" + name + "').");
    }
    
    public void plPortPropertySet(String name) {
	writer.println("portpropertyset('" + name + "').");
    }

    public void plComponentPropertySet(String name) {
	writer.println("componentpropertyset('" + name + "').");
    }

    public void plDataType(String name) {
	writer.println("datatype('" + name + "').");
    }

    public void plBuiltInType(String name) {
	writer.println("builtintype('" + name + "').");
    }

    public void plBaseType(String compName, String name) {
	writer.println("basetype('" + compName + "', '" + name + "').");
    }

    public void plField(String compName, String type, String[] dims, String name) {
	writer.println("field('" + compName + "', '" + type + "', [" + makeStringList(dims) + "], '" + name + "').");
    }
    

    private static String makeAtom(String str) {
	return(new StringBuffer("'").append(str).append("'").toString());
    }

    private static String makeStringList(String[] args) {
	StringBuffer res = new StringBuffer();

	for (int i = 0; i < args.length; i++) {
	    res.append("'");
	    res.append(args[i]);
	    res.append("'");
	    if (i < args.length - 1) {res.append(", ");}
	}
	return(res.toString());
    }

    private static String removeQuotes(String str) {
	if (str.startsWith("\"") && str.endsWith("\"")) {
	    return(str.substring(1, str.length() - 1));
	} else {
	    return(str);
	}
    }

}
