package coco.gui;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.awt.event.*;
import java.awt.*;

/**
 * This class implements a simple dialog for prompting the user for
 * some input.
 * Usage:
 * In the simplest form, just use the NameDialog this way:
 * String name = NameDialog.showNameDialog(parent, "Give name");
 * If the result is != null, the user clicked 'OK' and you get his text.
 *
 * To provide special labels for the input fields, use this form:
 * String name = NameDialog.showNameDialog(parent, "Give name", "Label A", "Label B");
 *
 * To provide checking, completing and error algorithms, you can use this:
 * String name = NameDialog.showNameDialog(parent, "Give name", "Label A", "Label B", new NameDialog.DialogHelperAdapter() {
 *
 *         public String complete(String name) {return(name);}
 *         public boolean isValid(String name) {return(true);}
 *         public void onError(JDialog dialog, String name, String file) {
 * 	     // display error window
 *	     JOptionPane.showMessageDialog(dialog, "The name " + name + " is invalid.", "Invalid name", JOptionPane.ERROR_MESSAGE);
 *	   }
 * });
 *
 * You do'nt have to provide all three methods: just overwrite the one, you want to change. Example: the name is only valid, if it contains a "-":
 * String name = NameDialog.showNameDialog(parent, "Give name", "Label A", "Label B", new NameDialog.DialogHelperAdapter() {
 *
 *         public boolean isValid(String name) {return(name.indexOf("-") > -1);}
 * });
 */

public class NameDialog extends JDialog implements DocumentListener {
    JTextField name;
    JTextField file;

    boolean success;

    DialogHelper helper;

    /**
     * Constructor: takes parent and title
     */
    public NameDialog(Frame parent, String title) {
	this(parent, title, "Name of the new component", "Filename");
    }

    /**
     * Constructor: takes parent, title, label for the name and label for the file input
     */
    public NameDialog(Frame parent, String title, String nameLabel, String fileLabel) {
	this(parent, title, nameLabel, fileLabel, new DialogHelperAdapter());
    }

    /**
     * Constructor: takes parent, title, label for the name and label for the file input
     */
    public NameDialog(Frame parent, String title, String nameLabel, String fileLabel, DialogHelper helper) {
	super(parent, title, true);
	init(nameLabel, fileLabel);
	this.helper = helper;
    }

    private void init(String nameLabel, String fileLabel) {
	name = new JTextField();
	name.setEditable(true);
	name.getDocument().addDocumentListener(this);

	file = new JTextField();
	file.setEditable(false);

	success = false;

	JPanel top = new JPanel();
	top.setBorder(new EmptyBorder(5, 5, 5, 5));
	top.setLayout(new GridLayout(4, 1));
	top.add(new JLabel(nameLabel));
	top.add(name);
	top.add(new JLabel(fileLabel));
	top.add(file);

	JPanel bottom = new JPanel();
	bottom.setBorder(new EmptyBorder(5, 5, 5, 5));
	bottom.setLayout(new GridLayout(1, 2));

	JButton b1 = new JButton("Ok");
	b1.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    b1_actionPerformed(e);
		}
	    });
	bottom.add(b1);

	JButton b2 = new JButton("Cancel");
	b2.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    b2_actionPerformed(e);
		}
	    });
	bottom.add(b2);

	getContentPane().setLayout(new BorderLayout());
	getContentPane().add(top, BorderLayout.CENTER);
	getContentPane().add(bottom, BorderLayout.SOUTH);
        b1.setSelected(true);
        b1.requestDefaultFocus();

        this.addKeyListener(new KeyAdapter(){
          public void keyPressed(KeyEvent e){
            if(e.getKeyCode() == e.VK_ENTER)
              handleOk();
            else if(e.getKeyCode() == e.VK_ESCAPE)
              handleCancel();
          }

        }


        );

        this.center();
	pack();
    }


    /**
     * Checks, if the input from the user is ok.
     * Here, the filename check must be performed...
     */
    private boolean checkInput() {
	if (helper != null) {
	    return(helper.isValid(getName()));
	} else {
	    return(true);
	}
    }


    // Call back methods for the buttons
    /**
     * The 'Ok' method
     */
    private void b1_actionPerformed(ActionEvent e) {
      handleOk();
    }

    private void handleOk(){
	if (checkInput()) {
	    success = true;
	    dispose();

	} else {
	    if (helper != null) {
		helper.onError(this, getName(), getFileName());
	    } else {
		// display error window
		JOptionPane.showMessageDialog(this, "The name " + getName() + " is invalid.", "Invalid name", JOptionPane.ERROR_MESSAGE);
	    }
	}
    }

    /**
     * The 'Cancel' method
     */
    private void b2_actionPerformed(ActionEvent e) {
      handleCancel();
    }

    private void handleCancel(){
	success = false;
	dispose();

    }


    // Methods for retrieving the input
    /**
     * Retrieves the component name
     */
    public String getName() {
	return(name.getText());
    }

    /**
     * Retrieves the file name
     */
    public String getFileName() {
	return(file.getText());
    }


    /**
     * Methods from the DocumentListener interface
     */
    public void changedUpdate(DocumentEvent e) {
	updateText();
    }

    /**
     * Methods from the DocumentListener interface
     */
    public void insertUpdate(DocumentEvent e) {
	updateText();
    }

    /**
     * Methods from the DocumentListener interface
     */
    public void removeUpdate(DocumentEvent e) {
	updateText();
    }

    /**
     * This method updates the second text field
     * Here the conversion from component name to file name must be performed
     */
    private void updateText() {
	if (helper != null) {
	    file.setText(helper.complete(getName()));
	} else {
	    file.setText(getName());
	}
    }

    /**
     * Did the user press the OK button?
     * Was the input valid?
     */
    public boolean success() {
	return(success);
    }

    /**
     * This is the static calling method of this dialog.
     * @param parent Parent frame of the dialog; can be null
     * @param title Title string of the dialog
     * @return Name of the new component, if the user pressed ok, or null
     */
    public static String showNameDialog(Frame parent, String title, String nameLabel, String fileLabel, DialogHelper helper) {
	return(showDialog(new NameDialog(parent, title, nameLabel, fileLabel, helper)));
    }

    /**
     * This is the static calling method of this dialog.
     * @param parent Parent frame of the dialog; can be null
     * @param title Title string of the dialog
     * @return Name of the new component, if the user pressed ok, or null
     */
    public static String showNameDialog(Frame parent, String title, String nameLabel, String fileLabel) {
	return(showDialog(new NameDialog(parent, title, nameLabel, fileLabel)));
    }

    /**
     * This is the static calling method of this dialog.
     * @param parent Parent frame of the dialog; can be null
     * @param title Title string of the dialog
     * @return Name of the new component, if the user pressed ok, or null
     */
    public static String showNameDialog(Frame parent, String title) {
	return(showDialog(new NameDialog(parent, title)));
    }

    /**
     * Internal dialog show method.
     */
    private static String showDialog(NameDialog d) {
	String res = null;

	if (d != null) {
	    d.show();

	    if (d.success()) {
		res = d.getName();
	    }
	}
	return(res);
    }

  void center(){
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

    Dimension frameSize = this.getSize();

    if (frameSize.height > screenSize.height) {
      frameSize.height = screenSize.height;
    }
    if (frameSize.width > screenSize.width) {
      frameSize.width = screenSize.width;
    }
    this.setLocation((screenSize.width - frameSize.width) / 2, (screenSize.height - frameSize.height) / 2);

  }



    /**
     * This interface provides support for functions of this dialog.
     */
    public interface DialogHelper {
	/**
	 * This is called to calculate the content of the second input line from the user input.
	 */
	public String complete(String name);

	/**
	 * This method is called, when the user clicks 'OK' to confirm his input.
	 */
	public boolean isValid(String name);

	/**
	 * This method is called, when the user clicked 'OK', but the input is invalid.
	 */
	public void onError(JDialog dialog, String name, String file);
    }

    /**
     * This class implements the DialogHelper interface with some simple behaviour.
     * The user just has to overwrite some of these methods, according to its needs.
     */
    public static class DialogHelperAdapter implements DialogHelper {/**
	 * Return the user input as file name.
	 */
	public String complete(String name) {return(name+".acm");}

	/**
	 * Input is always vaild.
	 */
	public boolean isValid(String name) {return(true);}

	/**
	 * In case of an invalid input, just display an error window.
	 */
	public void onError(JDialog dialog, String name, String file) {
	    // display error window
	    JOptionPane.showMessageDialog(dialog, "The name " + name + " is invalid.", "Invalid name", JOptionPane.ERROR_MESSAGE);
	}
    }
}
