package coco.gui.editor;

import java.awt.*;
import javax.swing.*;
import java.io.*;
import java.util.*;

/** 
 * This class implements a file editor which can edit multiple files at one time. 
 * Files are identified by their file names. To open a file for editing, use the 'openFile(...)' method.
 * This method opens the file and creates a new tab for the tabbed pane, using the supplied title.
 * After opening files, you can identify the editor windows by their file names. So the method 'getTitle(File f)'
 * retrieves the tab title of the editor tab, which edits the file f.
 * You can give an editor tab a new name for its file: just use the 'renameFile(...)' method. Afterwards use 
 * 'saveFile(...)' to use this new name for actually writing the file.
 * The 'getSelectedFile()' method retrieves the file name of the editor, which is currently active.
 * Beside file editor windows, MultiEdit can also provide simple display tabs. These tabs are usually used for 
 * error or status messages. These tabs are identified by the name of their tab title. The 'addDisplayTab(...)' 
 * method creates a new display tab. 'getDisplayTabs()' returns all display tab titles. Via 'get/setDisplayTabText(...)'
 * you can get/set the contents of the tab. To use these tabs efficiently, you can use 'getDisplayTabOutputStream(...)'
 * to get an OutputStream, which writes directly into the named display tab. You can also specify to rise the display tab,
 * whenever something is written to it. Display tabs are usually not editable.
 * This example demonstrates the usage of MultiEdit.
 * <pre>
 *   public static void main(String[] args) {
 *	JFrame fr = new JFrame("Testframe");
 *
 *	final MultiEdit me = new MultiEdit();
 *
 *      // Two display tabs for errors and messages
 *	me.addDisplayTab("Errors");
 *	me.addDisplayTab("Messages");
 *
 *	// get printstream for these displaytabs; both should use AutoRise
 *	final PrintStream errStream = new PrintStream(me.getDisplayTabOutputStream("Errors", true));
 *	final PrintStream msgStream = new PrintStream(me.getDisplayTabOutputStream("Messages", true));
 *
 *	fr.getContentPane().setLayout(new BorderLayout());
 *	fr.getContentPane().add(me, BorderLayout.CENTER);
 *
 *  	JButton b1 = new JButton("Open");
 * 	b1.addActionListener(new ActionListener() {
 * 		public void actionPerformed(ActionEvent e) {
 *
 *		    JFileChooser fc = new JFileChooser();
 *		    if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
 *			File f = fc.getSelectedFile();
 *
 *			if (f != null) {
 *			    msgStream.println("Open: " + f);
 *			    me.openFile(f, f.getName());
 *			}
 *		    }
 * 		}
 * 	    });
 *
 * 	JButton b2 = new JButton("Close");
 * 	b2.addActionListener(new ActionListener() {
 * 		public void actionPerformed(ActionEvent e) {
 *		    File f = me.getSelectedFile();
 *		    errStream.println(f);
 *		    me.closeFile(f);
 * 		}
 * 	    });
 *
 *	JPanel bp = new JPanel();
 *	bp.setLayout(new FlowLayout());
 *
 *	bp.add(b1);
 *	bp.add(b2);
 *
 *	fr.getContentPane().add(bp, BorderLayout.SOUTH);
 *
 *	fr.pack();
 *	fr.setVisible(true);
 *
 *    }
 * </pre>
 */
public class MultiEdit extends JPanel {

    // GUI
    JTabbedPane tab;

    Hashtable displayTabs;

    public MultiEdit() {
	super();
	init();
    }

    /**
     * Adds a new editor tab. The tab is identified by its name. Its content is set to the content parameter.
     * @param title Name of the new editor tab
     * @param content Initial content of the new editor tab
     */
    public void addEditorTab(String content, String title) {
	if (content != null) {
	    
	    StringEditor editor = new StringEditor(content, title);
	    tab.addTab(title, editor);
	}
    }

    /**
     * Returns the editor, specified by its title.
     * @param title Name of the editor
     */
    public StringEditor getEditorByTitle(String title) {
	StringEditor res = null;

	int cnt = tab.getTabCount();
	for (int i = 0; i < cnt; i++) {

	    Component cmp = tab.getComponentAt(i);

	    if (cmp instanceof StringEditor) {
		StringEditor editor = (StringEditor)cmp;
	    
		String t = editor.getTitle();

		if ((t != null) && (t.equals(title))) {
		    res = editor;
		    break;
		}
	    }
	}

	return(res);
    }

    /**
     * Removes the tab with the specified editor.
     */
    public void removeEditorTab(String title) {
	StringEditor editor = getEditorByTitle(title);
	tab.remove(editor);
    }

    /**
     * Retrieves an enumeration of all editors
     */
    public Enumeration getEditors() {

	Vector res = new Vector();

	int cnt = tab.getTabCount();
	for (int i = 0; i < cnt; i++) {
	    Component cmp = tab.getComponentAt(i);

	    if (cmp instanceof StringEditor) {
		StringEditor editor = (StringEditor)cmp;

		// just return the normal string editors, not the display tabs
		if (editor.getTitle() != null) {
		    res.add(editor.getTitle());
		}
	    }
	}

	return(res.elements());
    }

    /**
     * Retrieves the currently active editor title
     */
    public String getSelectedEditor() {
	String res = null;

	Component cmp = tab.getSelectedComponent();

	if (cmp instanceof StringEditor) {
	    StringEditor editor = (StringEditor)cmp;

	    if (editor != null) {
		res = editor.getTitle();
	    }
	}
	
	return(res);
    }

    /**
     * Get the text of the named editor, as it actually is.
     * @param title Name of the editor
     */
    public String getText(String title) {
	String res = null;

	StringEditor editor = getEditorByTitle(title);
	res = editor.getText();

	return(res);
    }

    /**
     * Reset the changed flag of the given editor.
     * @param title Name of the editor
     */
    public void resetChanged(String title) {
	StringEditor editor = getEditorByTitle(title);
	editor.resetChanged();
    }


    /**
     * Did the user edit in the named editor ?
     * @param title Name of the editor
     */
    public boolean hasChanged(String title) {
	boolean res = false;

	StringEditor editor = getEditorByTitle(title);
	res = editor.hasChanged();

	return(res);
    }

    /**
     * Did the user edit in the named editor, since the last request to this method?
     * @param title Name of the editor
     */
    public boolean hasFurtherChanged(String title) {
	boolean res = false;

	StringEditor editor = getEditorByTitle(title);
	res = editor.hasFurtherChanged();

	return(res);
    }

    /**
     * Set the caret of the named editor to the given line.
     * @param title Name of the editor
     * @param line Line to set the caret to
     */
    public void goToLine(String title, int line) {
	StringEditor editor = getEditorByTitle(title);
	editor.goToLine(line);
    }

    /**
     * Selects the named editor tab
     * @param title Name of the editor
     */
    public void selectEditor(String title) {
	StringEditor editor = getEditorByTitle(title);
	tab.setSelectedComponent(editor);
    }





    // Display tabs
    // Display tabs represent just displays without file representation and editing capabilities,
    // like error windows, ...

    /**
     * Adds a new display tab. Display tabs are identified by their names.
     */
    public void addDisplayTab(String title) {

	if (!displayTabs.containsKey(title)) {
	    PlainEditor editor = new PlainEditor(title);
	    editor.setEditable(false);

	    displayTabs.put(title, editor);

	    tab.addTab(title, editor);
	}
    }
    
    /**
     * Removes the named display tab.
     */
    public void removeDisplayTab(String title) {

	PlainEditor ed = (PlainEditor)displayTabs.get(title);

	if (ed != null) {

	    int cnt = tab.getTabCount();
	    int delAt = cnt;
	    for (int i = 0; i < cnt; i++) {
		Component cmp = tab.getComponentAt(i);

		if (cmp instanceof PlainEditor) {
		    PlainEditor editor = (PlainEditor)cmp;

		    if (editor == ed) {
			delAt = i;
			break;
		    }
		}
	    }

	    if (delAt < cnt) {
		displayTabs.remove(title);
		tab.removeTabAt(delAt);
	    }
	}
    }

    /**
     * Retrieves the names of all display tabs.
     */
    public Enumeration getDisplayTabs() {
	return(displayTabs.keys());
    }

    /**
     * Selects the named siaply tab to be active.
     */
    public void selectDisplayTab(String title) {

	PlainEditor ed = (PlainEditor)displayTabs.get(title);
	
	if (ed != null) {
	    tab.setSelectedComponent(ed);
	}
    }

    /**
     * Retrieves the text of the named display tab.
     */
    public String getDisplayTabText(String title) {
	String res = null;

	PlainEditor ed = (PlainEditor)displayTabs.get(title);
	
	if (ed != null) {
	    res = ed.getText();
	}

	return(res);
    }

    /**
     * Sets the text of the named display tab.
     */
    public void setDisplayTabText(String title, String text) {

	PlainEditor ed = (PlainEditor)displayTabs.get(title);
	
	if (ed != null) {
	    ed.setText(text);
	}
    }

    /** 
     * Retrieves an output stream for the named display tab.
     * This is useful for redirecting error or debugging messages to display tabs.
     */
    public OutputStream getDisplayTabOutputStream(String title, boolean autoRise) {
	return(new DTOutputStream(this, title, autoRise));
    }

    /**
     * This class represents an output stream for a display tab.
     * It can be used for printing to display tabs.
     */
    class DTOutputStream extends OutputStream {

	MultiEdit editor;
	String displayTab;
	boolean autoRise;

	public DTOutputStream(MultiEdit e, String dTab, boolean rise) {
	    super();

	    editor = e;
	    displayTab = dTab;
	    autoRise = rise;
	}
	
	public void write(int b) {
	    StringBuffer content = new StringBuffer(editor.getDisplayTabText(displayTab));
	    content.append((char)b);
	    editor.setDisplayTabText(displayTab, content.toString());
	    
	    if (autoRise) {
		editor.selectDisplayTab(displayTab);
	    }
	}
    }
   

    void init() {
	setLayout(new BorderLayout());

	tab = new JTabbedPane(JTabbedPane.BOTTOM);
	add(tab, BorderLayout.CENTER);

	displayTabs = new Hashtable();

    }
}


