package coco.gui.editor;

import org.gjt.sp.jedit.textarea.*;
import org.gjt.sp.jedit.syntax.*;
import coco.config.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

import coco.model.Controller;

/**
 * This class implements a simple string editor, using the JEditTextArea component.
 * It is a JPanel, containing the editor area. To edit a string, use the constructor.
 *
 */

public class StringEditor extends JPanel {

    String initialContent;
    String title;

    JEditTextArea textArea;
    JLabel statusLine;

    // this flag keeps the state relative to the initial content of the editor
    boolean dirty;

    // this flag keeps the state relative to the last 'isDirty(...)' request
    boolean tempDirty;

    public StringEditor(String content, String title) {
	super();

	initialContent = content;
	this.title = title;

	init();
	initText();
    }

    void init() {
	setLayout(new BorderLayout());

	// settings
	TextAreaDefaults defaults = TextAreaDefaults.getDefaults();
	defaults.eolMarkers = false;

	textArea = new JEditTextArea(defaults);

	// token manager
	textArea.setTokenMarker(new CoCoTokenMarker());

	// key bindings
	InputHandler handler = new DefaultInputHandler();
	handler.addDefaultKeyBindings();

	handler.addKeyBinding("C+c", new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    textArea.copy();
		}
	    });

	handler.addKeyBinding("C+x", new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    textArea.cut();
		}
	    });

	handler.addKeyBinding("C+v", new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    textArea.paste();
		}
	    });

	handler.addKeyBinding("S+INSERT", new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    textArea.paste();
		}
	    });

	handler.addKeyBinding("C+INSERT", new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    textArea.copy();
		}
	    });

	handler.addKeyBinding("S+DELETE", new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    textArea.cut();
		}
	    });



	handler.addKeyBinding("C+e", InputHandler.SELECT_DOC_END);
	handler.addKeyBinding("C+a", InputHandler.SELECT_DOC_HOME);

	handler.addKeyBinding("C+LEFT", InputHandler.PREV_WORD);
	handler.addKeyBinding("C+RIGHT", InputHandler.NEXT_WORD);

        // save file
	handler.addKeyBinding("C+s", new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // call the controller to save the current file.
                Controller controller = Controller.getInstance();
                String name = getTitle();
                System.out.println("Saving file " + name);
                controller.saveFile(name);
            }
          });

        handler.addKeyBinding("C+F4", new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Controller controller = Controller.getInstance();
                String name = getTitle();
                System.out.println("Closing Editor Tab " + name);
                controller.closeTab(name);
            }
          });


	textArea.setInputHandler(handler);

	add(textArea, BorderLayout.CENTER);

	statusLine = new JLabel("Position");
	add(statusLine, BorderLayout.SOUTH);

	textArea.addCaretListener(new CaretListener() {
		public void caretUpdate(CaretEvent e) {updateStatusLine();}
	    });

	textArea.getDocument().addDocumentListener(new DocumentListener() {
		public void changedUpdate(DocumentEvent e) {update();}
		public void insertUpdate(DocumentEvent e) {update();}
		public void removeUpdate(DocumentEvent e) {update();}

		private void update() {
		    dirty = true;
		    tempDirty = true;
		}

	    });

	// setting the tabsize
	GlobalConfig cfg = GlobalConfig.getInstance();
	String tabSize = cfg.getProperty(GlobalConfig.OPTION_EDITOR_TABSIZE);

	int ts = 2;
	if (tabSize != null) {
	    try {
		ts = Integer.parseInt(tabSize);
	    } catch (Exception ex) {}
	}

	// tab size
	textArea.getDocument().putProperty(PlainDocument.tabSizeAttribute, new Integer(ts));

    }

    /**
     * Set the content of the text area to the given initial content
     */
    void initText() {
	if (initialContent != null) {
	    textArea.setText(initialContent);

	    dirty = false;
	    tempDirty = false;
	}
    }

    /**
     * This method updates the status line of the window.
     */
    private void updateStatusLine() {
	int line = textArea.getCaretLine();
	int col = textArea.getCaretPosition() - textArea.getLineStartOffset(line);

	if (hasChanged()) {
	    statusLine.setText((line+1) + ":" + (col+1) + " *");
	} else {
	    statusLine.setText((line+1) + ":" + (col+1));
	}
    }

    public String getTitle() {
	return(title);
    }

    public void setTitle(String s) {
	title = s;
    }

    /**
     * Resets the changed flag. Useful when text has been saved.
     */
    public void resetChanged() {
	dirty = false;
	updateStatusLine();
    }

    public boolean hasChanged() {
	return(dirty);
    }

    public boolean hasFurtherChanged() {
	boolean res = tempDirty;
	tempDirty = false;
	return(res);
    }

    public void setEditable(boolean e) {
	textArea.setEditable(e);
    }

    public String getText() {
	return(textArea.getText());
    }

    public void setText(String s) {
	textArea.setText(s);
    }

    public void goToLine(int line) {
	textArea.scrollTo(line, 0);
	int pos = textArea.getLineStartOffset(line);
	textArea.setCaretPosition(pos);
    }

    public int getLineCount() {
	return(textArea.getLineCount());
    }

}
