package coco.parser.ast;

import java.util.Enumeration;

import coco.parser.symtab.*;
import coco.parser.ParseException;
import coco.utils.*;
import coco.parser.ast.enums.IdentifierEnumeration;

/**
 * berschrift:
 * Beschreibung:
 * Copyright:     Copyright (c) 2001
 * Organisation:
 * @author
 * @version 1.0
 */

public class ASTComponent extends GenericNode implements ASTComponentType, ASTPropertyContainer{

  public static final String UNIT_TYPE = "component";

  private java.io.File sourceFile = null;
  protected ASTIdentifier componentId;
  protected ASTInterfaceIdentifierList interfaceList = null;
  protected boolean activeComponent = false;
  protected ASTPropertySetReferenceList propSets = null;
  protected ASTComponentBody body = null;


  protected ASTComponent(){}

 public ASTComponent(ASTIdentifier i, boolean active, ASTInterfaceIdentifierList iList,  ASTPropertySetReferenceList ps, ASTComponentBody b){

    Assert.require(i);
    componentId = i;

    interfaceList = iList;
    if(interfaceList != null){
     interfaceList.setParent(this);
    }
    activeComponent = active;
    propSets = ps;
    if(propSets != null)
      propSets.setParent(this);
    Assert.require(b);
    body = b;
   // if (body != null)
    body.setParent(this);
  }

  public static GenericNode parse(String text) throws coco.parser.ParseException {
  Assert.require(text);
  coco.parser.CoCoParser parse = coco.parser.ParserService.getCoCoParser(new java.io.StringReader(text));
  ASTComponent c = parse.component();
  return c;
  }


  /**
   * Creates a template for a compilation unit with the given name
   *
   * @param name   the name of the unit
   * @return the newly created compilation unit template
   */
  public static ASTCompilationUnit createTemplateFromName(String name) throws coco.parser.ParseException{
    String text = "component " + name +"  {description: \"This is a new component\" } ";
    coco.parser.CoCoParser parser = coco.parser.ParserService.getCoCoParser(new java.io.StringReader(text));
    return parser.compilationUnit();
  }

  /**
   * Creates a template for component with the given name
   *
   * @param name   the name of the unit
   * @return the newly created compilation unit template
   */
  public static ASTComponent createNew(String name) throws SymtabDuplicateEntryException{
    Assert.require(name);
    ASTIdentifier newId = new ASTIdentifier(name);
    ASTComponentBody newBody = new ASTComponentBody(null, null, null, null);
    ASTComponent res = new ASTComponent(newId, true, null, null, newBody);
    //res.addProperty((ASTProperty)ASTProperty.createNew("newProperty"));
    // register
    res.register();
    return res;
  }



  public void addPort(ASTPort p) throws SymtabDuplicateEntryException{
    Assert.require(p);
    SymtabService.registerLocal(this, SymtabService.S_PORT, p.getPortId().getID(), p);

    ASTPortList pl = body.getPortList();
    if(pl == null){
      pl = new ASTPortList(p);
      body.setPortList(pl);
    }else{
      pl.addNode(p);
    }
  }

  public void removePort(ASTPort p){
    Assert.require(p);
    // there must be a port list, if somethings is gonna be removed.
    ASTPortList pl = body.getPortList();
    Assert.require(pl);

    if(SymtabService.lookupLocal(this, SymtabService.S_PORT, p.getPortId().getID()) != null){
      SymtabService.remove(this, SymtabService.S_PORT, p.getPortId().getID());
    }
    pl.removeNode(p);
  }

   public void addProperty(ASTProperty p) throws SymtabDuplicateEntryException{

    Assert.require(p);
    Assert.require(body);

    SymtabService.registerLocal(this, SymtabService.S_PROPERTY, p.getPropertyId().getID(), p);

    ASTPropertyList pl = body.getPropertyList();
    if(pl != null)
      pl.addNode(p);
    else{
      pl = new ASTPropertyList(p);
      body.setPropertyList(pl);
    }
  }

  public void removeProperty(ASTProperty p){
    // there must be a property list, if somethings is gonna be removed.
    ASTPropertyList pl = body.getPropertyList();
    Assert.require(pl);

    if(SymtabService.lookupLocal(this, SymtabService.S_PROPERTY, p.getPropertyId().getID()) != null){
      SymtabService.remove(this, SymtabService.S_PROPERTY, p.getPropertyId().getID());
    }
    pl.removeNode(p);

  }

   public ASTPropertyList getPropertyList(){
    Assert.require(body);
    return body.getPropertyList();
  }



  public String getUnitName(){
    return getID().toString()+".cm";
  }

  public String getUnitType(){
      return(UNIT_TYPE);
  }



  // registers local symbols too
  public void register() throws SymtabDuplicateEntryException{

      SymtabService.register(this);

      // register local symbols
      // ports
      ASTPortList pl = body.getPortList();
      if (pl != null) {
        for (Enumeration enum = pl.getNodes(); enum.hasMoreElements(); ) {
          ASTPort p = (ASTPort)enum.nextElement();
          SymtabService.registerLocal(this, SymtabService.S_PORT, p.getPortId().getID(), p);
        }
      }
      // properties
      ASTPropertyList properties = body.getPropertyList();
      if (properties != null) {

        for (Enumeration enum = properties.getNodes(); enum.hasMoreElements(); ) {
            ASTProperty p = (ASTProperty)enum.nextElement();
            SymtabService.registerLocal(this, SymtabService.S_PROPERTY, p.getPropertyId().getID(), p);
        }
      }


  }


  public boolean isActive(){
    return activeComponent;
  }

  public void setActive(boolean active){
    activeComponent = active;
  }


  /**
   * Copy all data from another generic node into this generic node
   * resync the syntax tree. The parameter src must not be null and must be of type ASTComponent
   *
   * @param src   the source node
   */
  public void copyFrom(GenericNode src) throws SymtabDuplicateEntryException{
    super.copyFrom(src);
    // unregister from the symbol table
    this.unregister();
    Assert.require(src instanceof ASTComponent);

    ASTComponent srcComp = (ASTComponent) src;

    this.setId(srcComp.getID());

    this.setInterfaceList(srcComp.getInterfaceList());
    this.setBody(srcComp.getBody());
    this.setActive(srcComp.isActive());


    // only take the source file info if there is one. otherwise keep the existing one
    if (srcComp.getSourceFile() != null)
      this.setSourceFile(srcComp.getSourceFile());


    // establish the back link from the body node
    body.setParent(this);

    // re-register the component
    this.register();

  }


/**
 * This function removes the astComponent entry from the symbol table.
 */

  public void unregister() {
  SymtabService.remove(this);
   }


  public void setSourceFile(java.io.File f){
    Assert.require(f);
    sourceFile = f;
  }
  public java.io.File getSourceFile(){
    return sourceFile;
  }




  public ASTIdentifierList getInterfaceList(){
      return interfaceList;
  }

  /**
   * This is not a necessary parameter of a component. So the parameter to this function ASTIdentifier list
   * can be null. That is why there is no need for require .
   */
  public void setInterfaceList(ASTIdentifierList il) {

    Assert.require(il instanceof ASTInterfaceIdentifierList);
    this.interfaceList = (ASTInterfaceIdentifierList)il;
    if (interfaceList != null) this.interfaceList.setParent(this);
    }




  public ASTIdentifier getID(){
    return componentId;
  }

  public void setId(ASTIdentifier identifier) {
  Assert.require(identifier);
  this.componentId = identifier;
  }

  public ASTIdentifier getContainerID(){
    return getID();
  }
  public String getName(){
    Assert.require(componentId);
    return componentId.getID();
  }

  public ASTComponentTypeBody getBody(){
    return body;
  }

  /**
   * setBody sets the body of ASTComponent. This component is set to be the parent of body.
   * @param body  astComponentTypeBody
   */

  public void setBody(ASTComponentTypeBody body) {
  Assert.require(body);
  this.body = (ASTComponentBody) body;
  body.setParent(this);
  }


  public String toString(){

    String res = "";
    if(beforeComments != null)
      res = beforeComments.toString();

    res += "component " + componentId;
    if(interfaceList != null)
      res += " is " + interfaceList.toString();

    if(propSets != null)
      res +=" has " + propSets.toString();
    if(body != null)
      res += body.toString();

    if(afterComments != null)
      res += afterComments.toString();

    return res;
  }


  public String toPrologString(){
    Assert.require(componentId);


    String res = "component(";

    IdentifierEnumeration enum;
    ASTIdentifier id;

    res += componentId.toPrologString();
    res += ").\n";


    if(interfaceList != null){
      enum = (IdentifierEnumeration) interfaceList.getNodes();
      while(enum.hasMoreElements()){
        id = (ASTIdentifier)enum.nextElement();
        res += "implements(" + componentId.toPrologString() + ", " +
          id.toPrologString() + ").\n";
      }
    }
    if(body != null)
      res += body.toPrologString();
    return res;
  }


  public Object accept(AbstractASTVisitor v, Object data) {
    if(v.preVisit()){
      v.visitPre(this);
    }
    // visit the sub nodes
    Assert.require(componentId);
    componentId.accept(v, data);

    if(interfaceList != null)
      interfaceList.accept(v, data);

    if(propSets != null)
      propSets.accept(v, data);


    Assert.require(body);
    body.accept(v, data);


    if(v.postVisit()){
      v.visitPost(this);
    }
    return null;
  }

}
