package coco.parser.ast;
import java.util.*;
import coco.parser.symtab.*;

import coco.parser.ParseException;
import coco.utils.*;
import coco.parser.ast.enums.IdentifierEnumeration;


/**
 * berschrift:   PECOS Composition Environment
 * Beschreibung:
 * Copyright:     Copyright (c) 2001
 * Organisation:  FZI
 * @author T. Genssler, A. Christoph
 * @version 1.0
 */

public class ASTInterface extends GenericNode implements ASTComponentType, ASTPropertyContainer{

  public static final String UNIT_TYPE = "interface";
  private ASTPropertySetReferenceList propSets = null;

  private java.io.File sourceFile = null;
  ASTIdentifier interfaceId;
  ASTIdentifierList interfaceList = null;
  ASTInterfaceBody body = null;

  public static GenericNode parse(String text) throws coco.parser.ParseException {
    Assert.require(text);
    coco.parser.CoCoParser parse = coco.parser.ParserService.getCoCoParser(new java.io.StringReader(text));
    ASTInterface c = parse.abstractComponent();
    return c;
  }

 /**
   * Copy all data from another generic node into this generic node
   * resync the syntax tree. The parameter src must not be null and must be of type ASTComponent
   *
   * @param src   the source node
   */
  public void copyFrom(GenericNode src) throws SymtabDuplicateEntryException{
    super.copyFrom(src);
    // unregister from the symbol table
    this.unregister();
    Assert.require(src instanceof ASTInterface);
    ASTInterface srcInterface = (ASTInterface) src;


    this.setId(srcInterface.getID());

    this.setInterfaceList(srcInterface.getInterfaceList());
    this.setBody(srcInterface.getBody());

    // only take the source file info if there is one. otherwise keep the existing one
    if (srcInterface.getSourceFile() != null)
      this.setSourceFile(srcInterface.getSourceFile());

    // establish the back link from the body node
    body.setParent(this);

    // re-register the component
    this.register();

    //catch(SymtabDuplicateEntryException error) {error.printStackTrace();}

  }


 /**
   * Creates a template for a compilation unit with the given name
   *
   * @param name   the name of the unit
   * @return the newly created compilation unit template
   */
  public static ASTCompilationUnit createTemplateFromName(String name) throws coco.parser.ParseException{
    String text = "interface " + name + "{ports: in data int newport; description: \"This is a new interface\"}";
    coco.parser.CoCoParser parser = coco.parser.ParserService.getCoCoParser(new java.io.StringReader(text));
    return parser.compilationUnit();
  }





  public String getUnitName(){
    return getID().getID()+".cm";
  }
  public String getUnitType(){
      return(UNIT_TYPE);
  }

  public ASTInterface(ASTIdentifier i, ASTIdentifierList iList, ASTPropertySetReferenceList ps, ASTInterfaceBody b) throws ParseException{

    Assert.require(i);
    interfaceId = i;
    interfaceList = iList;

    propSets = ps;
    if(propSets != null)
      propSets.setParent(this);
    Assert.require(b);
    body = b;
    body.setParent(this);
  }

  public void register() throws SymtabDuplicateEntryException{

      SymtabService.register(this);

      // register local symbols

      if (body!=null) {
        // ports
        ASTPortList pl = body.getPortList();
        if (pl != null) {
          for (Enumeration enum = pl.getNodes(); enum.hasMoreElements(); ) {
            ASTPort p = (ASTPort)enum.nextElement();
            SymtabService.registerLocal(this, SymtabService.S_PORT, p.getPortId().getID(), p);
          }

        }
        // properties
        ASTPropertyList properties = body.getPropertyList();
        if(properties != null){
          for (Enumeration enum = properties.getNodes(); enum.hasMoreElements(); ) {
              ASTAbstractProperty p = (ASTAbstractProperty)enum.nextElement();
              SymtabService.registerLocal(this, SymtabService.S_PROPERTY, p.getPropertyId().getID(), p);
          }
        }

	// register roles
	ASTRoleList roles = body.getRoles();
	if (roles != null) {
	    for (Enumeration enum = roles.getNodes(); enum.hasMoreElements(); ) {
		ASTRole role = (ASTRole)enum.nextElement();
		SymtabService.registerLocal(this, SymtabService.S_ROLE, role.getVarId().getID(), role);
	    }
	}

	// register instances
	ASTInstanceList instances = body.getInstances();
	if (instances != null) {
	    for (Enumeration enum = instances.getNodes(); enum.hasMoreElements(); ) {
		ASTInstance instance = (ASTInstance)enum.nextElement();
		SymtabService.registerLocal(this, SymtabService.S_INSTANCES, instance.getVarId().getID(), instance);
	    }
	}

	// register connectors
	ASTConnectorList connectors = body.getConnectors();
	if (connectors != null) {
	    for (Enumeration enum = connectors.getNodes(); enum.hasMoreElements(); ) {
		ASTConnector connector = (ASTConnector)enum.nextElement();
		SymtabService.registerLocal(this, SymtabService.S_CONNECTOR, connector.getConnectorID().getID(), connector);
	    }
	}
    }
  }

  public void unregister() {
  SymtabService.remove(this);
  }

  public ASTIdentifierList getInterfaceList(){
      return interfaceList;
  }

  public void setInterfaceList(ASTIdentifierList interfaceList) {
  this.interfaceList = interfaceList;
  }

  public void setSourceFile(java.io.File f){
    Assert.require(f);
    sourceFile = f;
  }
  public java.io.File getSourceFile(){
    return sourceFile;
  }


  public ASTIdentifier getID(){
    return interfaceId;
  }

  /**
   *  setId sets the identifier.
   *  Assertion is made here because this should not be null.
   */

  public void setId(ASTIdentifier identifier) {
  Assert.require(identifier);
  this.interfaceId = identifier;
  }

  public ASTIdentifier getContainerID(){
    return getID();
  }
  public String getName(){
    return interfaceId.toString();
  }

  public ASTComponentTypeBody getBody(){
    return body;
  }

  public void setBody(ASTComponentTypeBody body) {
  this.body = (ASTInterfaceBody)body;
  if (body!= null) body.setParent(this);
  }

  public String toString(){
    String res = "";

    res += "abstract component " + interfaceId;
    if(interfaceList != null)
      res+= " extends " + interfaceList.toString();
   if(propSets != null)
      res +=" has " + propSets.toString();

    if(body != null)
      res += body.toString();
    return res;
  }
  public String toPrologString(){
    Assert.require(interfaceId);
    String res = "interface(";

    res += interfaceId.toPrologString() +").\n";
    if(interfaceList != null){
      IdentifierEnumeration enum = (IdentifierEnumeration)interfaceList.getNodes();
      ASTIdentifier id = null;
      while(enum.hasMoreElements()){
        id = (ASTIdentifier) enum.nextElement();
        res += "implements(" + interfaceId.toPrologString() + ", " +
          id.toPrologString() + ").\n";
      }
    }
    if(body != null);
      res += body.toPrologString();

    return res;
  }



  public Object accept(AbstractASTVisitor v, Object data) {
    if(v.preVisit()){
      v.visitPre(this);
    }
    // visit the sub nodes
    Assert.require(interfaceId);
    interfaceId.accept(v, data);

    if(interfaceList != null)
      interfaceList.accept(v, data);
    if(propSets != null)
      propSets.accept(v, data);

    if(body != null);
      body.accept(v, data);


    if(v.postVisit()){
      v.visitPost(this);
    }
    return null;
  }

  public void addPort(ASTPort p) throws SymtabDuplicateEntryException{
    Assert.require(p);
    SymtabService.registerLocal(this, SymtabService.S_PORT, p.getPortId().getID(), p);

    ASTPortList pl = body.getPortList();
    if(pl != null){
      pl = new ASTPortList(p);
      body.setPortList(pl);
    }else{
      pl.addNode(p);
    }
  }

  public void removePort(ASTPort p){
    Assert.require(p);
    // there must be a port list, if somethings is gonna be removed.
    ASTPortList pl = body.getPortList();
    Assert.require(pl);
    if(SymtabService.lookupLocal(this, SymtabService.S_PORT, p.getPortId().getID()) != null){
      SymtabService.remove(this, SymtabService.S_PORT, p.getPortId().getID());
    }
    pl.removeNode(p);

  }



 public void addProperty(ASTProperty p) throws SymtabDuplicateEntryException{
    Assert.require(p);
    Assert.require(body);
    Assert.require(p instanceof ASTAbstractProperty);

    SymtabService.registerLocal(this, SymtabService.S_PROPERTY, p.getPropertyId().getID(), p);

    ASTAbstractPropertyList pl = (ASTAbstractPropertyList)body.getPropertyList();
    if(pl != null)
      pl.addNode(p);
    else{
      pl = new ASTAbstractPropertyList((ASTAbstractProperty)p);
      body.setPropertyList(pl);
    }
  }

  public void removeProperty(ASTProperty p){
    Assert.require(p instanceof ASTAbstractProperty);
    // there must be a property list, if somethings is gonna be removed.
    ASTPropertyList pl = body.getPropertyList();
    Assert.require(pl);

    if(SymtabService.lookupLocal(this, SymtabService.S_PROPERTY, p.getPropertyId().getID()) != null){
      SymtabService.remove(this, SymtabService.S_PROPERTY, p.getPropertyId().getID());
    }
    pl.removeNode(p);

  }


   public ASTPropertyList getPropertyList(){
    if(body != null)
      return body.getPropertyList();
    else
      return null;
  }


}
