package coco.semantic;

import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;


/**
 * This class tests ASTComponents.
 */
public class ComponentTester extends Tester {

    ASTComponent myComp;

    /**
     * Checks, if this tester can test the given node.
     * @parameter n Node to test
     * @return true: Tester can test the given node
     */
    protected boolean isCompatible(GenericNode n) {
	boolean res = false;

	if (n instanceof ASTComponent) {
	    res = true;
	}
	return(res);
    }

    /**
     * The 'real' test method. Is called, if and only if, the isCompatible method
     * returned 'true'.
     * @parameter n node to test
     * @return true: test succeeded
     */
    protected boolean testInternal(GenericNode n) {
	boolean res = true;

	ASTComponent c = (ASTComponent)n;
	ASTComponentBody b = (ASTComponentBody)c.getBody();

	myComp = c;

	// all implemented interfaces known ?
	// all interfaces properly implemented ?
	boolean intf_decl = true;

	ASTIdentifierList intf_list = checkList(c.getInterfaceList());

	for (Enumeration e = checkEnum(intf_list.getNodes()); e.hasMoreElements(); ) {
	    ASTIdentifier id = (ASTIdentifier)e.nextElement();

	    if (!(isInterfaceDeclared(id))) {
		intf_decl = false;
		reportError("unknown interface " + id);

	    } else {

		ASTInterface intf = lookupInterface(id);
		Assert.require(intf);

		if (!doesImplement(c, intf)) {
		    res = false;
		    reportError("interface " + id + " not properly implemented");
		}
	    }
	}

	// all used types known ?
	// all ports must use known data types or generic types

	// check ports
	// get ports from symtab service with subject = S_PORT
	Enumeration enum = checkEnum(SymtabService.getLocals(c, SymtabService.S_PORT));
	if (enum != null) {
	    for (; enum.hasMoreElements(); ) {
		ASTPort p = (ASTPort)enum.nextElement();

		if (!checkPortType(p)) {
		    res = false;
		}
	    }
	}

	// check generic types
	enum = SymtabService.getLocals(c, SymtabService.S_TYPEPARAM);
	if (enum != null) {
	    for (; enum.hasMoreElements(); ) {
		ASTGenericTypeParameter p = (ASTGenericTypeParameter)enum.nextElement();

		// all generic types must be interfaces
		if (!(isInterfaceDeclared(p.getTypeID()))) {
		    res = false;
		    reportError("unknown generic type " + p);
		}
	    }
	}

	// if composition exists, check its used types
	// and its semantics

	if (b != null ) {
	    ASTCompositionBody comp = b.getCompositionBody();

	    if (comp != null) {
		if (checkCompositionTypes(comp)) {
		    if (!checkCompositionSemantics(c, comp)) {
			res = false;
		    }
		} else {
		    res = false;
		}
	    }
	}

	return(res);
    }

    protected void reportError(String msg) {
	String ctx = "ComponentTester";

	if (myComp != null) {
	    ctx = ctx + " (" + myComp.getID() + ")";
	}

	reportError(ctx, msg);
    }

    /**
     * Checks, if the given component implements the given interface.
     * @param comp Component to be checked
     * @param intf Interface
     */
    private boolean doesImplement(ASTComponent comp, ASTInterface intf) {
	// are all ports of intf implemented in comp and have the same type ?

	boolean res = true;

	Assert.require(comp);
	Assert.require(intf);

	Enumeration intfPorts = checkEnum(SymtabService.getLocals(intf, SymtabService.S_PORT));

	// interface empty: we are right
	// interface != empty: comp must not be empty and match...
	if (intfPorts != null) {
	    for ( ; intfPorts.hasMoreElements(); ) {
		ASTPort intfp = (ASTPort)intfPorts.nextElement();
		String portId = getPortID(intfp).getID();
		ASTPort compp = (ASTPort)SymtabService.lookupLocal(comp, SymtabService.S_PORT, portId);

		if (compp != null) {
		    if (!compare(compp, intfp)) {
			res = false;
			reportError("interface port " + portId + " and component port " + getPortID(compp) + " do not match");
		    }
		} else {
		    res = false;
		    reportError("component " + comp.getID() + " does not implement port " + portId + " of interface " + intf.getID());
		}
	    }
	}

	return(res);
    }


    /**
     * Checks, if the symbols used by the composition are already known
     */
    private boolean checkCompositionTypes(ASTCompositionBody comp) {
	boolean res = true;
	ASTComponent parent = (ASTComponent)comp.getParent();

	//instances
	// gibt es alle verwendeten Connectoren/Componenten ?
	// Tabelle anlegen, mit lokaler Name -> Komponente


	// try to resolve instances list
	// i.e. try to find a component for every instance ind the
	// instance list
	// store this mapping in SymtabService, Subject 'Instances'
	ASTInstanceList il = comp.getInstances();
	if (il != null) {
	    for (Enumeration enum = checkEnum(il.getNodes()); enum.hasMoreElements(); ) {
		ASTInstance in = (ASTInstance)enum.nextElement();

		String typeId = in.getTypeId().getID();
		String varId = in.getVarId().getID();

		ASTComponentType localComp = (ASTComponentType)SymtabService.lookupComponent(typeId);

		if (localComp == null) {
		    ASTGenericTypeParameter tp = (ASTGenericTypeParameter)SymtabService.lookupLocal(parent, SymtabService.S_TYPEPARAM, typeId);

		    if (tp != null) {
			localComp = (ASTComponentType)SymtabService.lookupInterface(tp.getTypeID().getID());
		    }
		}

		if (localComp != null) {
		    try {
			SymtabService.registerLocal(parent, SymtabService.S_INSTANCES, varId, localComp);
		    } catch (Exception ex) {
			ex.printStackTrace();
		    }
		} else {

		    res = false;
		    reportError("component not found: " + typeId);
		    //          break;
		}
	    }
	}

	// wires
	// gibt es die referenzierten Ports ?
	// gibt es die Connectors
	ASTWireList wl = comp.getWires();

	if (wl != null) {
	    for (Enumeration enum = checkEnum(wl.getNodes()); enum.hasMoreElements(); ) {
		ASTWire w = (ASTWire)enum.nextElement();

		ASTDesignator des = w.getDesignator();
		String conId = des.toString();

		// does connector exist ?
		if (SymtabService.lookupConnector(conId) == null) {
		    res = false;
		    reportError("connector not found: " + conId);
		    //          break;
		}

		// existieren die Ports ?
		ASTPortAccessList sourcePorts = w.getSourcePorts();

		if (!checkPortAccessList(sourcePorts, parent)) {
		    res = false;
		    //          break;
		}

		ASTPortAccessList targetPorts = w.getTargetPorts();

		if (!checkPortAccessList(targetPorts, parent)) {
		    res = false;
		    //          break;
		}
	    }
	}

	return(res);
    }

    /**
     * Checks, if the ports addressed by the list exist
     */
    private boolean checkPortAccessList(ASTPortAccessList l, ASTComponent parent) {
	boolean res = true;

	if ((l != null) && (parent != null)) {
	    for (Enumeration enum = checkEnum(l.getNodes()); enum.hasMoreElements(); ) {
		ASTPortAccess portAccess = (ASTPortAccess)enum.nextElement();
		ASTDesignator des = portAccess.getDesignator();

		String compName = ((ASTIdentifier)des.getNodes().nextElement()).getID();

		// get the component referenced by the first element of the designator
		ASTComponentType comp = null;
		if (compName.equals("this")) {
		    comp = parent;
		} else {
		    comp = (ASTComponentType)SymtabService.lookupLocal(parent, SymtabService.S_INSTANCES, compName);
		}

		// if comp exists: check, if comp containes second part of the designator
		if (comp != null) {
		    Enumeration dummy = checkEnum(des.getNodes());
		    dummy.nextElement();
		    String portid = ((ASTIdentifier)dummy.nextElement()).getID();

		    // TODO: nicht nur nach Namen, auch nach Signatur bei Verwendung von Ports schauen
		    // BSP: zugriff auf t1.execute(int i), tc ist vom Typ C1, dann muss es einen execute(int) in C1 geben !
		    // sonst Exception !!

		    ASTPort port = null;
		    if (comp instanceof ASTComponent) {
			port = (ASTPort)SymtabService.lookupLocal((ASTComponent)comp, SymtabService.S_PORT, portid);

		    } else if (comp instanceof ASTInterface) {

			port = (ASTPort)SymtabService.lookupLocal((ASTInterface)comp, SymtabService.S_PORT, portid);
		    }

		    if (port == null) {
			res = false;
			reportError("port in component not found: " + portid);
		    } else {

			// registers dereferenced port with PortAccess
			// prerequisit for second stage of semantic checking
			try {
			    SymtabService.registerLocal(parent, SymtabService.S_PORTACCESS, portAccess.getDesignator().toString(), port);
			} catch (Exception ex) {
			    ex.printStackTrace();
			}
		    }

		} else {
		    res = false;
		    reportError("component not found: " + compName);
		}
	    }
	} else {
	    res = false;
	}


	return(res);


    }


    /**
     * Checks wires
     * @param parent Component containing the composition
     * @param body Composition body
     */
    private boolean checkCompositionSemantics(ASTComponent parent, ASTCompositionBody body) {
	boolean res = true;

	if (body != null) {
	    ASTWireList wl = body.getWires();

	    if (wl != null) {
		for (Enumeration enum = checkEnum(wl.getNodes()); enum.hasMoreElements(); ) {
		    ASTWire w = (ASTWire)enum.nextElement();

		    ASTDesignator des = w.getDesignator();
		    String conId = des.toString();
		    ASTConnector con = SymtabService.lookupConnector(conId);

		    // connector must exist, since we made previous analysis
		    Assert.require(con);

		    printDebugMsg("** con: " + con + ":" + con.getSourcePorts().getNumChildren());
		    printDebugMsg("** wire: " + w + ":" + w.getSourcePorts().getNumChildren());

		    if (!isInstance(parent, con.getSourcePorts(), w.getSourcePorts())) {
			res = false;
			reportError("source ports not compatible");
			break;
		    }

		    if (!isInstance(parent, con.getTargetPorts(), w.getTargetPorts())) {
			res = false;
			reportError("target ports not compatible");
			break;
		    }


		    // just check directions of datas here, since sytnax prohibits invalid use of
		    // handlers and invocations

		    // source ports must be out-datas
		    if (w.getSourcePorts() != null) {
			Enumeration enum2 = checkEnum(w.getSourcePorts().getNodes());
			if (enum2 != null) {
			    for (; enum2.hasMoreElements(); ) {
				ASTPortAccess pa = (ASTPortAccess)enum2.nextElement();

				Assert.require(pa);
				ASTPort port = (ASTPort)SymtabService.lookupLocal(parent, SymtabService.S_PORTACCESS, pa.getDesignator().toString());
				Assert.require(port);

				if (port instanceof ASTData) {
				    ASTData sig = (ASTData)port;
				    if(!(sig.getDirection().equals(ASTData.OUT_DIR))) {
					res = false;
					reportError("source datas must have out direction");
					break;
				    }
				}
			    }
			}
		    }


		    // target ports must be in-datas
		    if (w.getTargetPorts() != null) {
			Enumeration enum2 = checkEnum(w.getTargetPorts().getNodes());
			if (enum2 != null) {
			    for (; enum2.hasMoreElements(); ) {
				ASTPortAccess pa = (ASTPortAccess)enum2.nextElement();

				Assert.require(pa);
				ASTPort port = (ASTPort)SymtabService.lookupLocal(parent, SymtabService.S_PORTACCESS, pa.getDesignator().toString());
				Assert.require(port);

				if (port instanceof ASTData) {
				    ASTData sig = (ASTData)port;
				    if(!(sig.getDirection().equals(ASTData.IN_DIR))) {
					res = false;
					reportError("target datas must have in direction");
					break;
				    }
				}
			    }
			}
		    }
		}
	    }
	} else {
	    res = false;
	}

	return(res);
    }

    /**
     * Checks, if a given PortAccessList is an instance of a given PortTypeReferenceList
     * @param parent Component containing the actual composition
     * @param portTypeRefs port type references from connector specifictaion
     * @param portAccessList port access list  from wire specification
     */
    private boolean isInstance(ASTComponent parent, ASTPortTypeReferenceList portTypeRefs, ASTPortAccessList portAccessList) {
	boolean res = true;
	if ((portTypeRefs != null) && (portAccessList != null)) {


	    printDebugMsg("** Len: " + portTypeRefs.getNumChildren() + ":" + portAccessList.getNumChildren());




	    int accessCnt = 0;
	    int index = 0;

	    for (Enumeration enum = checkEnum(portTypeRefs.getNodes()); enum.hasMoreElements(); ) {
		ASTPortTypeReference ref = (ASTPortTypeReference)enum.nextElement();

		printDebugMsg("** index: " + index);



		if (index >= portAccessList.getNumChildren()) {
		    res = false;
		    reportError("more port type references than port accesses");
		    break;

		} else {
		    if (ref.isArrayOfPorts()) {
			for (int i = index; i < portAccessList.getNumChildren(); i++) {
			    ASTPortAccess pa = (ASTPortAccess)portAccessList.getChild(i);
			    Assert.require(pa);
			    ASTPort port = (ASTPort)SymtabService.lookupLocal(parent, SymtabService.S_PORTACCESS, pa.getDesignator().toString());
			    Assert.require(port);

			    if (!compare(ref, port, false)) {
				index = i;
				break;
			    } else {
				accessCnt++;
			    }
			}

		    } else {

			ASTPortAccess pa = (ASTPortAccess)portAccessList.getChild(index);
			Assert.require(pa);
			ASTPort port = (ASTPort)SymtabService.lookupLocal(parent, SymtabService.S_PORTACCESS, pa.getDesignator().toString());
			Assert.require(port);

			printDebugMsg("** test:: " + ref + ":" + pa.getDesignator() + ":" + pa.getInvocationParameters());
			printDebugMsg("** port:: " + port);

			// passen port und portAccess zusammen ?
			//			if (compare(port, pa)) {

			    if (compare(ref, port, false)) {
				index++;
				accessCnt++;

			    } else {
				res = false;
				reportError("type reference " + ref.toString() + " and port " + pa.getDesignator().toString() + " do not match");
				break;
			    }

			    //			} else {
			    //			    res = false;
			    //			    reportError("port " + port + " and access " + pa + " do not match");
			    //			    break;
			    //			}


		    }
		}
	    }

	    printDebugMsg("** test: " + accessCnt + ":" + portAccessList.getNumChildren());


	    if (res && (accessCnt < portAccessList.getNumChildren())) {
		res = false;
		reportError("there are more port accesses than declared connector parameters");
	    }

	} else {
	    res = false;
	}
	return(res);
    }

    /**
     * Compares a PortTypeReference with a Port
     * @param ref port type reference
     * @param port port to compare to
     * @return true, if reference matches the port; else false
     */
    private boolean compare(ASTPortTypeReference ref, ASTPort port, boolean report) {
	boolean res = true;

	printDebugMsg("compare: " + ref + ":" + port);


	if ((ref != null) && (port != null)) {

	    if ((ref instanceof ASTDataTypeReference) && (port instanceof ASTData)) {
		ASTDataTypeIdentifier p1 = ((ASTDataTypeReference)ref).getTypeId();
		ASTDataTypeIdentifier p2 = ((ASTData)port).getDataType();

		if (!isEqual(p1, p2, report)) {
		    res = false;
		    if (report) reportError("packet types " + p1.getTypeID().getID() + " and " + p2.getTypeID().getID() + " do not match");
		}

	    } else if ((ref instanceof ASTInvocationTypeReference) && (port instanceof ASTInvocation)) {
		ASTInvocationSignature sig1 = ((ASTInvocationTypeReference)ref).getSignature();
		ASTInvocationSignature sig2= ((ASTInvocation)port).getInvocationSig();

		if (!isEqual(sig1, sig2)) {
		    res = false;
		    if (report) reportError("signatures of " + sig1.getId() + " and " + sig2.getId() + " do not match");
		}

	    } else if ((ref instanceof ASTHandlerTypeReference) && (port instanceof ASTHandler)) {
		ASTInvocationSignature sig1 = ((ASTHandlerTypeReference)ref).getSignature();
		ASTInvocationSignature sig2= ((ASTHandler)port).getHandlerSig();

		if (!isEqual(sig1, sig2)) {
		    res = false;
		    if (report) reportError("signatures of " + sig1.getId() + " and " + sig2.getId() + " do not match");
		}

	    } else {
		res = false;
	    }
	} else {
	    res = false;
	}
	return(res);
    }

    /**
     * Retrieves the ID of a given port.
     * @param p Port, which ID is needed
     */

    private ASTIdentifier getPortID(ASTPort p) {
	ASTIdentifier res = null;

	if (p instanceof ASTData) {
	    ASTData sig = (ASTData)p;
	    res = sig.getPortId();

	} else if (p instanceof ASTInvocation) {
	    ASTInvocation e = (ASTInvocation)p;
	    res = e.getInvocationSig().getId();

	} else if (p instanceof ASTHandler) {
	    ASTHandler h = (ASTHandler)p;
	    res = h.getHandlerSig().getId();
	}

	return(res);
    }
}





