package coco.semantic;

import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;


/**
 * This class tests ASTInterfaces (=abstract components)
 */
public class InterfaceTester extends Tester {

    ASTInterface myIntf;

    /**
     * Checks, if this tester can test the given node.
     * @parameter n Node to test
     * @return true: Tester can test the given node
     */
    protected boolean isCompatible(GenericNode n) {
	boolean res = false;

	if (n instanceof ASTInterface) {
	    res = true;
	}
	return(res);
    }

    /**
     * The 'real' test method. Is called, if and only if, the isCompatible method
     * returned 'true'.
     * @parameter n node to test
     * @return true: test succeeded
     */
    protected boolean testInternal(GenericNode n) {
	boolean res = true;

	ASTInterface i = (ASTInterface)n;
	myIntf = i;

	// all base interfaces known ?
	// all port types known ?

	Enumeration e = checkEnum(checkList(i.getInterfaceList()).getNodes());
	while(e.hasMoreElements()) {
	    ASTIdentifier id = (ASTIdentifier)e.nextElement();
	    
	    if (!(isInterfaceDeclared(id))) {
		res = false;
		reportError("unknown base interface " + id);
	    }
	}
	

	// check ports
	// get ports from symtab service with subject = S_PORT
	e = checkEnum(SymtabService.getLocals(myIntf, SymtabService.S_PORT));
	if (e != null) {
	    for (; e.hasMoreElements(); ) {
		ASTPort p = (ASTPort)e.nextElement();

		if (!checkPortType(p)) {
		    res = false;
		    reportError("unknown data type " + p);
		}
	    }
	}

	// check roles
	// do all roles have interfaces as types?
	// are these interfaces already known?

	e = checkEnum(SymtabService.getLocals(myIntf, SymtabService.S_ROLE));
	while (e.hasMoreElements()) {
	    ASTRole role = (ASTRole)e.nextElement();
	    ASTIdentifier id = role.getTypeId();

	    if (!(isInterfaceDeclared(id))) {
		res = false;
		reportError("unknown type in role " + role);
	    }
	}

	// check instances
	// are all types components and already known?
	// are all used role names known?

	e = checkEnum(SymtabService.getLocals(myIntf, SymtabService.S_INSTANCES));
	while (e.hasMoreElements()) {
	    ASTInstance instance = (ASTInstance)e.nextElement();
	    ASTIdentifier id = instance.getTypeId();

	    if (!(isComponentDeclared(id))) {
		res = false;
		reportError("unknown type in instance " + instance);

	    } else {
		// check the role name
		ASTRoleIdentifier roleId = instance.getRoleId();

		if (roleId != null) {
		    ASTRole role = getRoleDefinition(myIntf, roleId);

		    if (role != null) {
			ASTIdentifier typeId = role.getTypeId();

			if (isInterfaceDeclared(typeId)) {
			    if (!(doesImplement(lookupComponent(id), typeId))) {
				reportError("component and role not compatible " + id);
				res = false;
			    }

			} else {
			    reportError("unknown interface in role " + role);
			    res = false;
			}

		    } else {
			reportError("unknown role " + roleId);
			res = false;
		    }
		}
	    }
	}

	// todo: check connectors
	//
	//

	// what to check here? In-Out is only important for a singel component.
	// Outside a component, port are simply common data spaces

	return(res);
    }


    protected void reportError(String msg) {
	String ctx = "InterfaceTester";

	if (myIntf != null) {
	    ctx = ctx + " (" + myIntf.getID() + ")";
	}

	reportError(ctx, msg);
    }

    // helpers

    /**
     * Checks, if the data type used by this port is known.
     * If it has a primitive type, return true. Else, use the symtab to 
     * lookup the data type.
     */
    private boolean checkPortType(ASTPort port) {
	boolean res = false;

	ASTDataTypeIdentifier dtid = port.getDataType();
	Assert.require(dtid);

	if (isUserDefinedDataType(dtid)) {
	    res = isDataTypeDeclared(dtid.getTypeID());
	} else {
	    res = isPrimitiveDataType(dtid);
	}

	return(res);
    }

    /**
     * Search the role with the given name in thegiven interface.
     */
    private ASTRole getRole(ASTInterface intf, ASTIdentifier id) {
	ASTRole res = null;
	
	if ((intf != null) && (id != null)) {

	    Enumeration e = checkEnum(SymtabService.getLocals(intf, SymtabService.S_ROLE));
	    while (e.hasMoreElements()) {
		ASTRole role = (ASTRole)e.nextElement();
		ASTIdentifier rid = role.getVarId();
	    
		if (id.equals(rid)) {
		    res = role;
		    break;
		}
	    }
	}
	return(res);
    }

    /**
     * Search the given role definition in the given interfaces and all of its base interfaces.
     */
    private ASTRole getRoleDefinition(ASTInterface intf, ASTIdentifier id) {
	ASTRole res = null;
	
	if ((intf != null) && (id != null)) {
	    
	    // search in the given interface
	    res = getRole(intf, id);

	    // and in its base interfaces
	    if (res == null) {
		Enumeration e = TypeAnalyzer.getBaseInterfaces(intf);

		while (e.hasMoreElements()) {
		    ASTIdentifier id2 = (ASTIdentifier)e.nextElement();
		    ASTInterface intf2 = SymtabService.lookupInterface(id2.getID());

		    if (intf2 != null) {
			res = getRole(intf2, id);
			if (res != null) {
			    break;
			}
		    }
		}
	    }
	}	    
	return(res);
    }

    /**
     * Checks, if the given component implements the given interface
     */

    // Test anhand der 'is' Deklarationen
    private boolean doesImplement(ASTComponent comp, ASTIdentifier id) {
	boolean res = false;

	if ((comp != null) && (id != null)) {
	    res = TypeAnalyzer.doesImplement(comp, id);
	}
	return(res);
    }

    /**
     * Extracts the first part of a designator (the component id)
     */
    private ASTIdentifier getComponentId(ASTDesignator des) {
	ASTIdentifier res = null;

	if (des != null) {
	    Enumeration e = checkEnum(des.getNodes());

	    if (e.hasMoreElements()) {
		res = (ASTIdentifier)e.nextElement();
	    }
	}
	return(res);
    }

    /**
     * Extracts the second part of the designator (the port id)
     */
    private ASTIdentifier getPortId(ASTDesignator des) {
	ASTIdentifier res = null;

	if (des != null) {
	    Enumeration e = checkEnum(des.getNodes());

	    if (e.hasMoreElements()) {
		e.nextElement();

		if (e.hasMoreElements()) {
		    res = (ASTIdentifier)e.nextElement();
		}
	    }
	}
	return(res);
    }
}

