package coco.semantic;

import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;


/**
 * This class tests ASTRules
 */
public class RuleTester extends Tester {

    ASTRule myRule;

    /**
     * Checks, if this tester can test the given node.
     * @parameter n Node to test
     * @return true: Tester can test the given node
     */
    protected boolean isCompatible(GenericNode n) {
	boolean res = false;

	if (n instanceof ASTRule) {
	    res = true;
	}
	return(res);
    }

    /**
     * The 'real' test method. Is called, if and only if, the isCompatible method
     * returned 'true'.
     * @parameter n node to test
     * @return true: test succeeded
     */
    protected boolean testInternal(GenericNode n) {
	boolean res = true;

	ASTRule r = (ASTRule)n;
	myRule = r;

	ASTTermList terms = r.getTerms();

	// must not be zero
	Assert.require(terms);

	Enumeration enum = checkEnum(terms.getNodes());

	// must not be null
	Assert.require(enum);

	for (; enum.hasMoreElements(); ) {
	    // all terms must be quantified compund terms

	    ASTTerm term = (ASTTerm)enum.nextElement();

	    if (term instanceof ASTQuantifierCompoundTerm) {
		if (!checkCompoundTerm(r, (ASTQuantifierCompoundTerm)term)) {
		    res = false;
		}

	    } else {
		res = false;
		reportError("quantifier compound term expected");
		break;
	    }

	}

	return(res);
    }

    private boolean checkTerm(ASTRule r, ASTTerm t) {
	boolean res = true;

	if ((r != null) && (t != null)) {

	    if (t instanceof ASTCompoundTerm) {
		ASTCompoundTerm ct = (ASTCompoundTerm)t;

		res = checkCompoundTerm(r, ct);

	    } else if (t instanceof ASTList) {
		ASTList l = (ASTList)t;

		ASTTerm head = l.getHead();
		ASTTermList terms = l.getTail();

		if (head != null) {
		    res = checkTerm(r, head);
		}

		if (terms != null) {
		    res = res && checkTermList(r, terms);
		}

	    } else if (t instanceof ASTTuple) {
		ASTTuple tu = (ASTTuple)t;
		ASTTermList terms = tu.getTerms();

		if (terms != null) {
		    res = checkTermList(r, terms);
		}

	    } else if (t instanceof ASTIdentifier) {
		ASTIdentifier id = (ASTIdentifier)t;

		Assert.require(id);

		// must be local parameter, since rule references are 'compound terms'

		if (SymtabService.lookupLocal(r, SymtabService.S_TYPEPARAM, id.getID()) == null) {
		    res = false;
		    reportError("unknown identifier (" + id +")");
		}
	    }

	} else {
	    res = false;
	}
	return(res);
    }

    private boolean checkTermList(ASTRule r, ASTTermList terms) {
	boolean res = true;

	if((r != null) && (terms != null)) {

	    if (terms.getNumChildren() > 0) {
		for (Enumeration e = checkEnum(terms.getNodes()); e.hasMoreElements(); ) {
		    ASTTerm term = (ASTTerm)e.nextElement();
		    if (!checkTerm(r, term)) {
			res = false;
			reportError("term " + term + " not valid");
			break;
		    }
		}
	    }
	} else {
	    res = false;
	}
	return(res);
    }

    private boolean checkCompoundTerm(ASTRule r, ASTCompoundTerm ct) {
	boolean res = true;

	if ((r != null) && (ct != null)) {
	    if (ct instanceof ASTComparisonTerm) {
		ASTComparisonTerm comp = (ASTComparisonTerm)ct;

		// operand1 must be local parameter
		ASTIdentifier id = comp.getOperand1();

		Assert.require(id);
		if (SymtabService.lookupLocal(r, SymtabService.S_TYPEPARAM, id.getID()) == null) {
		    res = false;
		    reportError("identifier " + id + " unknown");

		} else {
		    ASTTerm term = comp.getOperand2();
		    Assert.require(term);
		    if (!checkTerm(r, term)) {
			res = false;
		    }
		}

	    } else if (ct instanceof ASTAssignmentTerm) {
		ASTAssignmentTerm at = (ASTAssignmentTerm)ct;

		// target must be local parameter
		ASTIdentifier id = at.getTarget();

		Assert.require(id);
		if (SymtabService.lookupLocal(r, SymtabService.S_TYPEPARAM, id.getID()) == null) {
		    res = false;
		    reportError("identifier " + id + " unknown");

		} else {
		    ASTTerm term = at.getSource();
		    Assert.require(term);
		    if (!checkTerm(r, term)) {
			res = false;
		    }
		}

	    } else if (ct instanceof ASTArithmeticTerm) {
		ASTArithmeticTerm at = (ASTArithmeticTerm)ct;

		// operand1 must be local parameter
		ASTIdentifier id = at.getOperand1();

		Assert.require(id);
		if (SymtabService.lookupLocal(r, SymtabService.S_TYPEPARAM, id.getID()) == null) {
		    res = false;
		    reportError("identifier " + id + " unknown");

		} else {
		    ASTTerm term = at.getOperand2();
		    Assert.require(term);
		    if (!checkTerm(r, term)) {
			res = false;
		    }
		}

	    } else if (ct instanceof ASTPredicateTerm) {
		ASTPredicateTerm pt = (ASTPredicateTerm)ct;

		// predicate must exist (= rule?)
		ASTIdentifier id = pt.getPredicateID();

		Assert.require(id);

		if (SymtabService.lookupRule(id.getID()) == null) {
		    res = false;
		    reportError("predicate " + id + " unknown");

		} else {
		    ASTTermList params = pt.getArguments();

		    if (!checkTermList(r, params)) {
			res = false;
		    }
		}

	    } else if (ct instanceof ASTQuantifierCompoundTerm) {

		ASTQuantifierCompoundTerm qterm = (ASTQuantifierCompoundTerm)ct;

		// check quantifier

		ASTQuantifierExpression q = qterm.getQuantifier();

		if (q != null) {

		    printDebugMsg("** QTErm: " + q);

		    // identifier must be local symbol
		    String id = q.getID().getID();

		    if (SymtabService.lookupLocal(r, SymtabService.S_TYPEPARAM, id) == null) {
			res = false;
			reportError("unknown id in quantifier (" + id + ")");
		    }

		    // scope must be local symbol
		    String sc = q.getScope().getID();

		    ASTFormalRuleParameter tp;

		    if ((tp = (ASTFormalRuleParameter)SymtabService.lookupLocal(r, SymtabService.S_TYPEPARAM, sc)) != null) {
			// scope must be a list ?
			printDebugMsg("** TypeParameter: " + tp);
			printDebugMsg("**** TODO: modify RuleParser.jj to check parameter types");

		    } else {
			res = false;
			reportError("unknown scope in quantifier");
		    }
		}


		// check term

		ASTTerm t = qterm.getTerm();

		if (!checkTerm(r, t)) {
		    res = false;
		}
	    }

	} else {
	    res = false;
	}

	return(res);
    }

    protected void reportError(String msg) {
	String ctx = "RuleTester";

	if (myRule != null) {
	    ctx = ctx + " (" + myRule.getRuleID() + ")";
	}

	reportError(ctx, msg);
    }
}

