package coco.semantic;

import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;

/**
 * This class represents a test class for a AST entity.
 * Concrete subclasses must overwrite methods, to
 * test a specific entity.
 */
public abstract class Tester {

    // debug messages ?
    protected boolean DEBUG = true;

    protected Vector errors = new Vector();

    /**
     * This factory method creates a new tester depending on the type of the argument.
     * @parameter n Node to be tested
     * @return tester for n
     */
    public static Tester getTester(GenericNode n) {
	Tester res = null;

	if (n instanceof ASTDataType) {
	    res = new DataTypeTester();
	} else if (n instanceof ASTComponent) {
	    res = new ComponentTester();
	} else if (n instanceof ASTInterface) {
	    res = new InterfaceTester();
	    //	} else if (n instanceof ASTRule) {
	    //	    res = new RuleTester();
	}

	return(res);
    }



    /**
     * Starting method for the test.
     * @parameter n the entity to test
     * @return true: test succeeded
     */
    public boolean test(GenericNode n) {
	boolean res = false;

	errors = new Vector();

	if (isCompatible(n)) {
	    res = testInternal(n);
	}
	return(res);
    }

    /**
     * Checks, if this tester can test the given node.
     * @parameter n Node to test
     * @return true: Tester can test the given node
     */
    protected abstract boolean isCompatible(GenericNode n);

    /**
     * The 'real' test method. Is called, if and only if, the isCompatible method
     * returned 'true'.
     * @parameter n node to test
     * @return true: test succeeded
     */
    protected abstract boolean testInternal(GenericNode n);

    protected abstract void reportError(String msg);

    // error messages

    protected void reportError(String subject, String msg) {
	errors.add("[" + subject + "] " + msg);
    }

    public String[] getErrors() {
	Assert.require(errors);
	return((String[])errors.toArray(new String[0]));
    }

    // debug messages
    protected void printDebugMsg(String msg) {
	if (DEBUG) {
	    System.out.println(msg);
	}
    }

    // helpers

    /**
     * Checks, if the given identifier is a declared datatype in the symtab.
     */
    protected boolean isDataTypeDeclared(ASTIdentifier id) {
	boolean res = false;

	if ((id != null) && (id.getID() != null)) {
	    if (SymtabService.lookupDataType(id.getID()) != null) {
		res = true;
	    }
	}

	return(res);
    }

    /**
     * Checks, if the supplied data type is a primitive one.
     */
    protected boolean isPrimitiveDataType(ASTDataTypeIdentifier id) {
	return(id instanceof ASTBaseTypeIdentifier);
    }

    /**
     * Checks, if the supplied data type is user defined/
     */
    protected boolean isUserDefinedDataType(ASTDataTypeIdentifier id) {
	return(id instanceof ASTUserTypeIdentifier);
    }

    /**
     * Looks up a given datatype in the symtab
     */
    protected ASTDataType lookupDataType(ASTIdentifier id) {
	ASTDataType res = null;

	if (isDataTypeDeclared(id)) {
	    res = SymtabService.lookupDataType(id.getID());
	}

	return(res);
    }

    /**
     * Checks, if the given interface id is already known.
     */
    protected boolean isInterfaceDeclared(ASTIdentifier id) {
	boolean res = false;

	if ((id != null) && (id.getID() != null)) {
	    if (SymtabService.lookupInterface(id.getID()) != null) {
		res = true;
	    }
	}

	return(res);
    }

    /**
     * Retrieves the interface with the given id from the symtab
     */
    protected ASTInterface lookupInterface(ASTIdentifier id) {
	ASTInterface res = null;

	if (isInterfaceDeclared(id)) {
	    res = SymtabService.lookupInterface(id.getID());
	}

	return(res);
    }

    /**
     * Checks, if the given component id is already known.
     */
    protected boolean isComponentDeclared(ASTIdentifier id) {
	boolean res = false;

	if ((id != null) && (id.getID() != null)) {
	    if (SymtabService.lookupComponent(id.getID()) != null) {
		res = true;
	    }
	}

	return(res);
    }

    /**
     * Retrieves the component with the given id from the symtab
     */
    protected ASTComponent lookupComponent(ASTIdentifier id) {
	ASTComponent res = null;

	if (isComponentDeclared(id)) {
	    res = SymtabService.lookupComponent(id.getID());
	}

	return(res);
    }



    


    // These methods take care of empty enums or lists

    /**
     * This method checks Enumerations for beeing null. 
     * @param e Enumeration to check
     * @return e, if e != null; for e == null, an empty Enumeration is returned.
     */
    protected Enumeration checkEnum(Enumeration e) {
	if (e != null) {
	    return(e);
	} else {
	    return(new Enumeration() {
		    public boolean hasMoreElements() {return(false);}
		    public Object nextElement() {return(null);}
		});
	}
    }


    /**
     * This method checks Lists for beeing null. 
     * @param l List to check
     * @return l, for l != null; an empty list for l == null
     */
    protected ASTIdentifierList checkList(ASTIdentifierList l) {
	if (l != null) {
	    return(l);
	} else {
	    return(new ASTIdentifierList());
	}
    }

    /**
     * This method checks Lists for beeing null. 
     * @param l List to check
     * @return l, for l != null; an empty list for l == null
     */
    protected ASTRoleList checkList(ASTRoleList l) {
	if (l != null) {
	    return(l);
	} else {
	    return(new ASTRoleList(null));
	}
    }

    /**
     * This method checks Lists for beeing null. 
     * @param l List to check
     * @return l, for l != null; an empty list for l == null
     */
    protected ASTInstanceList checkList(ASTInstanceList l) {
	if (l != null) {
	    return(l);
	} else {
	    return(new ASTInstanceList(null));
	}
    }

}
