package coco.semantic;

import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;

/**
 * This class represents a test class for a AST entity.
 * Concrete subclasses must overwrite methods, to
 * test a specific entity.
 */
public abstract class Tester {

    // debug messages ?
    protected boolean DEBUG = true;

    protected Vector errors = new Vector();

    /**
     * This factory method creates a new tester depending on the type of the argument.
     * @parameter n Node to be tested
     * @return tester for n
     */
    public static Tester getTester(GenericNode n) {
	Tester res = null;

	if (n instanceof ASTDataType) {
	    res = new DataTypeTester();
	    //	} else if (n instanceof ASTComponent) {
	    //	    res = new ComponentTester();
	    //	} else if (n instanceof ASTInterface) {
	    //	    res = new InterfaceTester();
	    //	} else if (n instanceof ASTRule) {
	    //	    res = new RuleTester();
	}

	return(res);
    }



    /**
     * Starting method for the test.
     * @parameter n the entity to test
     * @return true: test succeeded
     */
    public boolean test(GenericNode n) {
	boolean res = false;

	errors = new Vector();

	if (isCompatible(n)) {
	    res = testInternal(n);
	}
	return(res);
    }

    /**
     * Checks, if this tester can test the given node.
     * @parameter n Node to test
     * @return true: Tester can test the given node
     */
    protected abstract boolean isCompatible(GenericNode n);

    /**
     * The 'real' test method. Is called, if and only if, the isCompatible method
     * returned 'true'.
     * @parameter n node to test
     * @return true: test succeeded
     */
    protected abstract boolean testInternal(GenericNode n);

    protected abstract void reportError(String msg);

    // error messages

    protected void reportError(String subject, String msg) {
	errors.add("[" + subject + "] " + msg);
    }

    public String[] getErrors() {
	Assert.require(errors);
	return((String[])errors.toArray(new String[0]));
    }

    // debug messages
    protected void printDebugMsg(String msg) {
	if (DEBUG) {
	    System.out.println(msg);
	}
    }

    // helpers

    protected boolean isDataTypeDeclared(ASTIdentifier id) {
	boolean res = false;

	if ((id != null) && (id.getID() != null)) {
	    if (SymtabService.lookupDataType(id.getID()) != null) {
		res = true;
	    }
	}

	return(res);
    }

    protected ASTDataType lookupDataType(ASTIdentifier id) {
	ASTDataType res = null;

	if (isDataTypeDeclared(id)) {
	    res = SymtabService.lookupDataType(id.getID());
	}

	return(res);
    }

    protected boolean isInterfaceDeclared(ASTIdentifier id) {
	boolean res = false;

	if ((id != null) && (id.getID() != null)) {
	    if (SymtabService.lookupInterface(id.getID()) != null) {
		res = true;
	    }
	}

	return(res);
    }

    protected ASTInterface lookupInterface(ASTIdentifier id) {
	ASTInterface res = null;

	if (isInterfaceDeclared(id)) {
	    res = SymtabService.lookupInterface(id.getID());
	}

	return(res);
    }

    protected boolean isComponentDeclared(ASTIdentifier id) {
	boolean res = false;

	if ((id != null) && (id.getID() != null)) {
	    if (SymtabService.lookupComponent(id.getID()) != null) {
		res = true;
	    }
	}

	return(res);
    }

    protected ASTComponent lookupComponent(ASTIdentifier id) {
	ASTComponent res = null;

	if (isComponentDeclared(id)) {
	    res = SymtabService.lookupComponent(id.getID());
	}

	return(res);
    }

    /**
     * Checks, if two ports match
     * @param p1 Port 1
     * @param p2 Port 2
     */
    protected boolean compare(ASTPort p1, ASTPort p2) {
	boolean res = true;

	if ((p1 != null) && (p2 != null)) {

	    // direction
	    if (!(p1.getDirection().equals(p2.getDirection()))) {
		res = false;
	    }

	    // modifier check ?

	    // data type check
	    if ((p1 instanceof ASTData) && (p2 instanceof ASTData)) {
		ASTData sig1 = (ASTData)p1;
		ASTData sig2 = (ASTData)p2;

		if (!isEqual(sig1.getDataType(), sig2.getDataType(), true)) {
		    res = false;
		    reportError("port types are not compatible");
		}

	    } else if ((p1 instanceof ASTHandler) && (p2 instanceof ASTHandler)) {
		ASTInvocationSignature sign1 = ((ASTHandler)p1).getHandlerSig();
		ASTInvocationSignature sign2 = ((ASTHandler)p2).getHandlerSig();

		// not only parameters, also return types must be checked... (todo)

		if ((sign1 != null) && (sign2 != null)) {
		    if (!isEqual(sign1, sign2)) {
			res = false;
			reportError("Signatures are incompatible");
		    }
		}

		
	    }
	} else {
	    res = false;
	}

	return(res);
    }

    /**
    * Checks, if a given port and a given portaccess match. These entities are presumed to have the same name
    * and reference the same entity.
    * @param port Port to be compared
    * @param access Access to this port (uasge of this port)
    */
    public boolean compare(ASTPort port, ASTPortAccess access) {
	boolean res = true;

	if ((port != null) && (access != null)) {

	    if (port instanceof ASTInvocation) {
		ASTInvocation invocation = (ASTInvocation)port;

		ASTInvocationParameterList l1 = invocation.getInvocationSig().getInvocationParams();
		ASTInvocationParameterList l2 = access.getInvocationParameters();

		if ((l1 != null) || (l2 != null)) {
		    res = isEqual(l1, l2);
		}
	    } else if (port instanceof ASTHandler) {
		ASTHandler handler = (ASTHandler)port;

		ASTInvocationParameterList l1 = handler.getHandlerSig().getInvocationParams();
		ASTInvocationParameterList l2 = access.getInvocationParameters();

		printDebugMsg("**handler sigs: " + l1 + ":" + l2);

		if ((l1 != null) || (l2 != null)) {
		    res = isEqual(l1, l2);
		}
	    }

	} else {
	    res = false;
	}
	return(res);
    }

    /**
    * Checks, if two parameterlists are equal.
    * @param l1 first parameter list
    * @param l2 second parameter list
    */
    public boolean isEqual(ASTInvocationParameterList l1, ASTInvocationParameterList l2) {
	boolean res = true;

	if ((l1 != null) && (l2 != null)) {
	    if (l1.getNumChildren() == l2.getNumChildren()) {
		Enumeration el1 = l1.getNodes();
		Enumeration el2 = l2.getNodes();

		while (el1.hasMoreElements()) {
		    ASTInvocationParameter param1 = (ASTInvocationParameter)el1.nextElement();
		    ASTInvocationParameter param2 = (ASTInvocationParameter)el2.nextElement();

		    if (!isEqual(param1.getParamType(), param2.getParamType(), true)) {
			res = false;
			reportError("parameter types do not match");
			break;
		    }
		}
	    } else {
		res = false;
		reportError("parameter lists have different length");
	    }
	} else if ((l1 == null) && (l2 == null)) {
	} else {
	    res = false;
	    reportError("parameter lists do not match");
	}
	return(res);
    }

    /**
     * Checks, if two signatures are equal
     * @param sig1 First signature
     * @param sig2 Second signature
     */
    public boolean isEqual(ASTInvocationSignature sig1, ASTInvocationSignature sig2) {
	boolean res = true;

	if ((sig1 != null) && (sig2 != null)) {
	    ASTInvocationParameterList l1 = sig1.getInvocationParams();
	    ASTInvocationParameterList l2 = sig2.getInvocationParams();

	    //      printDebugMsg("** list1: " + l1 + ":" + sig1);
	    //	    printDebugMsg("** list2: " + l2 + ":" + sig2);

	    res = isEqual(l1, l2);

	} else {
	    res = false;
	}
	return(res);
    }

    /**
     * Checks, it two packet types are compatible
     * @param p1 First packet type
     * @param p2 Second packet type
     */
    protected boolean isEqual(ASTDataTypeIdentifier p1, ASTDataTypeIdentifier p2, boolean report) {
	boolean res = true;

	// check only equality of names, because packets do not have
	// generic type parameters anymore

	if ((p1 != null) && (p2 != null)) {
	    if (!p1.getTypeID().getID().equals(p2.getTypeID().getID())) {
		res = false;
		if (report) reportError("packet types have different names");
	    }

	} else {
	    res = false;
	}
	return(res);
    }

    /**
     * Checks, if one packet type is a subclass of another paket type
     * @param p1 First packet type
     * @param p2 Second packet type
     * @return true: p1 is subclass of p2
     */
    protected boolean isSubtypeOf(ASTDataTypeIdentifier p1, ASTDataTypeIdentifier p2) {
	Assert.require(p1);
	Assert.require(p2);
	return(isSubtypeOf(p1.getTypeID(), p2.getTypeID()));
    }

    /**
     * Checks, if one packet type is a subclass of another paket type
     * @param p1 First packet type
     * @param p2 Second packet type
     * @return true: p1 is subclass of p2
     */
    protected boolean isSubtypeOf(ASTIdentifier p1, ASTIdentifier p2) {

	// Test: p2 : A, p1 : B, B extends A -> true
	//       p2 : A, p1 : C, C extends B, B extends A -> true

	boolean res = false;

	Assert.require(p1);
	Assert.require(p2);

	ASTDataType packet1 = lookupDataType(p1);
	ASTDataType packet2 = lookupDataType(p2);

	Assert.require(packet1);
	Assert.require(packet2);

	if (p1.getID().equals(p2.getID())) {
	    res = true;
	} else {


            ASTIdentifier bt = packet1.getBasetypeIdentifier();

	    if (bt != null) {

		// p2 basetype of p1 ?

		if (bt.getID().equals(p2.getID())) {
		    res = true;
		} else {

		    // basetype of p1 = subtype of p2 ?
		    res = isSubtypeOf(bt, p2);
		}
	    }
	}
	return(res);
    }

    /**
     * Checks, if two signatures are compatible
     * @param sig1 First signature
     * @param sig2 Second signature
     */


    // TG

    /*


    protected boolean isCompatible(ASTInvocationSignature sig1, ASTInvocationSignature sig2) {
	boolean res = true;

	if ((sig1 != null) && (sig2 != null)) {
	    ASTInvocationParameterList l1 = sig1.getInvocationParams();
	    ASTInvocationParameterList l2 = sig2.getInvocationParams();

	    printDebugMsg("** list1: " + l1 + ":" + sig1);
	    printDebugMsg("** list2: " + l2 + ":" + sig2);

	    res = isCompatible(l1, l2);

	} else {
	    res = false;
	}
	return(res);
    }
    */

    /**
     * Checks, if two parameterlists are compatible.
     * @param l1 first parameter list
     * @param l2 second parameter list
     */
    /*
    protected boolean isCompatible(ASTInvocationParameterList l1, ASTInvocationParameterList l2) {
	boolean res = true;

	if ((l1 != null) && (l2 != null)) {
	    if (l1.getNumChildren() == l2.getNumChildren()) {
		Enumeration el1 = l1.getNodes();
		Enumeration el2 = l2.getNodes();

		while (el1.hasMoreElements()) {
		    ASTInvocationParameter param1 = (ASTInvocationParameter)el1.nextElement();
		    ASTInvocationParameter param2 = (ASTInvocationParameter)el2.nextElement();

		    if (!isCompatible(param1.getParamType(), param2.getParamType(), true)) {
			res = false;
			reportError("parameter types do not match");
			break;
		    }
		}
	    } else {
		res = false;
		reportError("parameter lists have different length");
	    }
	} else if ((l1 == null) && (l2 == null)) {
	} else {
	    res = false;
	    reportError("parameter lists do not match");
	}
	return(res);
    }




    */




    /**
     * This method tests, if all packet types used in the port p are known
     * @param p ASTPort object to test
     * @return true: all types are known
     */
    protected boolean checkPortType(ASTPort p) {
	boolean res = true;

	if (p instanceof ASTData) {
	    ASTData data = (ASTData)p;

	    // can we find the type of the data ?
	    // result will be false, if the data has neither
	    // a primitive type, nor a type registered in the symbol table
	    if (!checkType(data.getDataType(), null)) {
		res = false;
		reportError("unknown data type " + data.getDataType());
	    }

	} else {

	    ASTInvocationParameterList paramList = null;

	    if (p instanceof ASTInvocation) {
		ASTInvocation invocation = (ASTInvocation)p;

		// get all parameters
		paramList = invocation.getInvocationSig().getInvocationParams();

	    } else if (p instanceof ASTHandler) {
		ASTHandler h = (ASTHandler)p;

		// get all parameters
		paramList = h.getHandlerSig().getInvocationParams();
	    }

	    if (paramList != null) {
		for (Enumeration e = paramList.getNodes(); e.hasMoreElements(); ) {
		    ASTInvocationParameter param = (ASTInvocationParameter)e.nextElement();
		    ASTDataTypeIdentifier t = param.getParamType();

		    // can we find the type of the data ?
		    // result will be false, if the data has neither
		    // a primitive type, nor a type registered in the symbol table
		    if (!checkType(t, null)) {
			res = false;
			reportError("unknown type in signature " + t);
		    }
		}
	    }
	}
	return(res);
    }

    /**
     * Checks, if the packet type is either primitive or registered in the
     * packet symbol table
     * This method checks instanciated packets.
     * @param type packet type to be checked
     * @return true: type was either primitive or registered in the symbol table, false: type was not found
     */
    protected boolean checkType(ASTDataTypeIdentifier type, Vector localTypeParams) {
	boolean res = false;

	// can we find the type of the data ?
	// result will be false, if the data has neither
	// a primitive type, nor a type registered in the symbol table

	// built-in type ?
	if (type.isBuiltInType()) {
	    return(true);
	}

	// local type (i.e. type parameter from instanciated packet)
	if (localTypeParams != null) {
	    if (localTypeParams.contains(type.getTypeID().getID())) {
		return(true);
	    }
	}

	// compound or defined by a typelib ?
	if (isDataTypeDeclared(type.getTypeID())) {
	    return(true);
	}

	return(res);
    }

    
    /**
     * This method checks Enumerations for beeing null. 
     * @param e Enumeration to check
     * @return e, if e != null; for e == null, an empty Enumeration is returned.
     */
    protected Enumeration checkEnum(Enumeration e) {
	if (e != null) {
	    return(e);
	} else {
	    return(new Enumeration() {
		    public boolean hasMoreElements() {return(false);}
		    public Object nextElement() {return(null);}
		});
	}
    }


    /**
     * This method checks Lists for beeing null. 
     * @param l List to check
     * @return l, for l != null; an empty list for l == null
     */
    protected ASTIdentifierList checkList(ASTIdentifierList l) {
	if (l != null) {
	    return(l);
	} else {
	    return(new ASTIdentifierList());
	}
    }

}



