package coco.semantic;

import coco.parser.*;
import coco.parser.ast.*;
import coco.parser.symtab.*;

import java.util.*;


/**
 * This class tracks type relationa among model entities. It uses the TypeGraph.
 */
public class TypeAnalyzer {

    // some constants
    private static final String INTERFACE = "I_";
    private static final String COMPONENT = "C_";
    private static final String DATATYPE = "D_";

    /**
     * Add an interface and all of its base interfaces to the type graph.
     */
    public static void addInterface(ASTInterface intf) {
	
	if (intf != null) {
	    String from = INTERFACE + intf.getID().getID();

	    // get children
	    ASTIdentifierList ids = intf.getInterfaceList();
	    if (ids != null) {
		Enumeration e = checkEnum(ids.getNodes());
		while (e.hasMoreElements()) {
		    ASTIdentifier to_i = (ASTIdentifier)e.nextElement();
		    ASTInterface to_intf = SymtabService.lookupInterface(to_i.getID());

		    if (to_intf != null) {
			String to = INTERFACE + to_i.getID();
			getTypeGraph().addRelation(from, to);

			addInterface(to_intf);
		    }
		}
	    }
	}
    }

    /** 
     * Add a component and all of its implemented interfaces to the type graph.
     */
    public static void addComponent(ASTComponent cmp) {

	if (cmp != null) {
	    String from = COMPONENT + cmp.getID().getID();

	    // get children
	    ASTIdentifierList ids = cmp.getInterfaceList();
	    if (ids != null) {
		Enumeration e = checkEnum(ids.getNodes());
		while (e.hasMoreElements()) {
		    ASTIdentifier to_i = (ASTIdentifier)e.nextElement();
		    ASTInterface to_intf = SymtabService.lookupInterface(to_i.getID());

		    if (to_intf != null) {
			String to = INTERFACE + to_i.getID();
			getTypeGraph().addRelation(from, to);

			addInterface(to_intf);
		    }
		}
	    }
	}
    }

    /**
     * Adds a data type and its base types to the analyzer.
     */
    public static void addDataType(ASTDataType dtype) {

	if (dtype != null) {
	    String from = DATATYPE + dtype.getDataTypeIdentifier().getID();

	    // get base type
	    ASTIdentifier btid = dtype.getBasetypeIdentifier();
	    if (btid != null) {
		ASTDataType btype = SymtabService.lookupDataType(btid.getID());
		
		if (btype != null) {
		    String to = DATATYPE + btid.getID();
		    getTypeGraph().addRelation(from, to);

		    addDataType(btype);
		}
	    }
	}
    }

    /**
     * Does the given component implement the given interface?
     */
    public static boolean doesImplement(ASTComponent cmp, ASTIdentifier intf) {
	boolean res = false;

	if ((cmp != null) && (intf != null)) {
	    String from = COMPONENT + cmp.getID().getID();
	    String to = INTERFACE + intf.getID();

	    res = getTypeGraph().haveRelation(from, to);
	}
	return(res);
    }

    /**
     * Does the given Interface inherit from the given base interface?
     */
    public static boolean doesExtend(ASTInterface sub, ASTIdentifier base) {
	boolean res = false;

	if ((sub != null) && (base != null)) {
	    String from = INTERFACE + sub.getID().getID();
	    String to = INTERFACE + base.getID();

	    res = getTypeGraph().haveRelation(from, to);
	}

	return(res);
    }

    /**
     * Is the given data type a subtype of the given base type?
     */
    public static boolean isSubtype(ASTDataType dtype, ASTIdentifier base) {
	boolean res = false;
	
	if ((dtype != null) && (base != null)) {
	    String from = DATATYPE + dtype.getDataTypeIdentifier().getID();
	    String to = DATATYPE + base.getID();

	    res = getTypeGraph().haveRelation(from, to);
	}

	return(res);
    }

    /**
     * Retrieves all implemented interfaces of the given component.
     * @return Enumeration of ASTIdentifiers
     */
    public static Enumeration getBaseInterfaces(ASTComponent cmp) {
	Vector res = new Vector();

	if (cmp != null) {
	    String id = COMPONENT + cmp.getID().getID();

	    Enumeration e = getTypeGraph().getChildren(id);
	    while(e.hasMoreElements()) {
		String c_id = (String)e.nextElement();

		if (c_id.startsWith(INTERFACE)) {
		    res.add(new ASTIdentifier(c_id.substring(INTERFACE.length())));
		}
	    }
	}
	return(res.elements());
    }

    /**
     * Retrieves all base interfaces of the given interface.
     * @return Enumeration of ASTIdentifiers
     */
    public static Enumeration getBaseInterfaces(ASTInterface intf) {
	Vector res = new Vector();

	if (intf != null) {
	    String id = INTERFACE + intf.getID().getID();

	    Enumeration e = getTypeGraph().getChildren(id);
	    while(e.hasMoreElements()) {
		String c_id = (String)e.nextElement();

		if (c_id.startsWith(INTERFACE)) {
		    res.add(new ASTIdentifier(c_id.substring(INTERFACE.length())));
		}
	    }
	}
	return(res.elements());
    }

    /**
     * Retrieves all base types of the given data type.
     * @return Enumeration of ASTIdentifiers
     */
    public static Enumeration getBaseTypes(ASTDataType dtype) {
	Vector res = new Vector();

	if (dtype != null) {
	    String id = DATATYPE + dtype.getDataTypeIdentifier().getID();

	    Enumeration e = getTypeGraph().getChildren(id);
	    while(e.hasMoreElements()) {
		String c_id = (String)e.nextElement();

		if (c_id.startsWith(DATATYPE)) {
		    res.add(new ASTIdentifier(c_id.substring(DATATYPE.length())));
		}
	    }
	}
	return(res.elements());
    }




















    /**
     * Resets the analyzer, i.e. forget all relations.
     */
    public static void reset() {
	getTypeGraph().reset();
    }
    






    /**
     * Get the TypeGraph singleton.
     */
    private static TypeGraph getTypeGraph() {
	return(TypeGraph.getInstance());
    }

    /**
     * Check the given enumeration for null ;-)
     */
    private static Enumeration checkEnum(Enumeration e) {
	if (e != null) {
	    return(e);
	} else {
	    return(new Enumeration() {
		    public boolean hasMoreElements() {return(false);}
		    public Object nextElement() {return(null);}
		});
	}
    }
}
