package coco.semantic;

import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;


/**
 * This class tests ASTComponents.
 */
public class ComponentTester extends Tester {

    ASTComponent myComp;

    /**
     * Checks, if this tester can test the given node.
     * @parameter n Node to test
     * @return true: Tester can test the given node
     */
    protected boolean isCompatible(GenericNode n) {
	boolean res = false;

	if (n instanceof ASTComponent) {
	    res = true;
	}
	return(res);
    }

    /**
     * The 'real' test method. Is called, if and only if, the isCompatible method
     * returned 'true'.
     * @parameter n node to test
     * @return true: test succeeded
     */
    protected boolean testInternal(GenericNode n) {
	boolean res = true;

	ASTComponent c = (ASTComponent)n;
	ASTComponentBody b = (ASTComponentBody)c.getBody();

	myComp = c;

	// all implemented interfaces known ?
	// all interfaces properly implemented ?
	ASTIdentifierList intf_list = checkList(c.getInterfaceList());

	for (Enumeration e = checkEnum(intf_list.getNodes()); e.hasMoreElements(); ) {
	    ASTIdentifier id = (ASTIdentifier)e.nextElement();

	    if (!(isInterfaceDeclared(id))) {
		res = false;
		reportError("unknown interface " + id);
		break;

	    } else {

		if (!doesImplement(c, id)) {
		    res = false;
		    reportError("interface " + id + " not properly implemented");
		}
	    }
	}

	// all roles bound ?
	if (!allRolesBound(c)) {
	    res = false;
	}



	// all used types known ?
	// all ports must use known data types or generic types

	// check ports
	// get ports from symtab service with subject = S_PORT
	Enumeration enum = checkEnum(SymtabService.getLocals(c, SymtabService.S_PORT));
	if (enum != null) {
	    for (; enum.hasMoreElements(); ) {
		ASTPort p = (ASTPort)enum.nextElement();

		if (!checkPortType(p)) {
		    res = false;
                    reportError("unknwon type: " + p.getDataType().getTypeID());
		}
	    }
	}



	// check instances
	// are all types components and already known?
	// are all used role names known?

	enum = checkEnum(SymtabService.getLocals(c, SymtabService.S_INSTANCES));
	while (enum.hasMoreElements()) {
	    ASTInstance instance = (ASTInstance)enum.nextElement();
	    ASTIdentifier id = instance.getTypeId();

	    if (!(isComponentDeclared(id))) {
		res = false;
		reportError("unknown type in instance " + instance);

	    } else {
		// check the role name
		ASTRoleIdentifier roleId = instance.getRoleId();

		if (roleId != null) {
		    ASTRole role = getRoleDefinition(c, roleId);

		    if (role != null) {
			ASTIdentifier typeId = role.getTypeId();

			if (isInterfaceDeclared(typeId)) {
			    if (!(doesImplement(lookupComponent(id), typeId))) {
				reportError("component and role not compatible " + id);
				res = false;
			    }

			} else {
			    reportError("unknown interface in role " + role);
			    res = false;
			}

		    } else {
			reportError("unknown role " + roleId);
			res = false;
		    }
		}
	    }
	}



	// check connectors
	// what to check (see InterfaceTester for further questions)

	return(res);
    }

    protected void reportError(String msg) {
	String ctx = "ComponentTester";

	if (myComp != null) {
	    ctx = ctx + " (" + myComp.getID() + ")";
	}

	reportError(ctx, msg);
    }


    // Helpers


    /**
     * Checks, if the given component implements the given interface.
     * Asks the TypeAnalyzer...
     *
     */

    // Test anhand der 'is' Deklarationen
    private boolean doesImplement(ASTComponent comp, ASTIdentifier id) {
	boolean res = false;

	if ((comp != null) && (id != null)) {
	    res = TypeAnalyzer.doesImplement(comp, id);
	}
	return(res);
    }

    // Test, ob alle (geerbten) Rollen gebunden sind...

    /**
     * This method checks, if all inherited roles are bound.
     */
    public boolean allRolesBound(ASTComponent comp) {
	boolean res = true;

	if (comp != null) {
	    // get all roles and instances
	    Enumeration roles = getAllRoles(comp);
	    Enumeration instances = getAllInstances(comp);

	    // for every role, check, if an instance exists
	    while (roles.hasMoreElements()) {
		ASTRole role = (ASTRole)roles.nextElement();

		ASTIdentifier roleId = role.getVarId();
		ASTIdentifier typeId = role.getTypeId();

		boolean found = false;

		while (instances.hasMoreElements()) {
		    ASTInstance inst = (ASTInstance)instances.nextElement();

		    ASTIdentifier rId = inst.getRoleId();
		    ASTComponent cmp = lookupComponent(inst.getTypeId());

		    if (rId != null) {
			if (rId.equals(roleId)) {
			    if (TypeAnalyzer.doesImplement(cmp, typeId)) {
				found = true;
			    }
			}
		    }
		}

		if (!found) {
		    res = false;
		    reportError("Role " + roleId + " not bound to an appropriate instance");
		    break;
		}
	    }
	}
	return(res);
    }

    /**
     * This method retrieves all roles of all base interfaces of the given component.
     * @return Enumeration of ASTRole
     */
    private Enumeration getAllRoles(ASTComponent comp) {
	Vector res = new Vector();

	if (comp != null) {
	    Enumeration intfs = checkEnum(TypeAnalyzer.getBaseInterfaces(comp));

	    while (intfs.hasMoreElements()) {
		ASTInterface intf = lookupInterface((ASTIdentifier)intfs.nextElement());
		Assert.require(intf);

		Enumeration roles = checkEnum(SymtabService.getLocals(intf, SymtabService.S_ROLE));

		while (roles.hasMoreElements()) {
		    ASTRole role = (ASTRole)roles.nextElement();
		    res.add(role);
		}
	    }
	}
	return(res.elements());
    }

    /**
     * This method retrieves all instances of a component and of all base interfaces of this component.
     * @return Enumeration of ASTInstance.
     */
    private Enumeration getAllInstances(ASTComponent comp) {
	Vector res = new Vector();

	if (comp != null) {
	    // add own instances
	    Enumeration instances = checkList(((ASTComponentBody)comp.getBody()).getInstances()).getNodes();

	    while (instances.hasMoreElements()) {
		ASTInstance inst = (ASTInstance)instances.nextElement();
		res.add(inst);
	    }

	    // add all instances of base interfaces
	    Enumeration intfs = checkEnum(TypeAnalyzer.getBaseInterfaces(comp));

	    while (intfs.hasMoreElements()) {
		ASTInterface intf = lookupInterface((ASTIdentifier)intfs.nextElement());
		Assert.require(intf);

		Enumeration insts = checkEnum(SymtabService.getLocals(intf, SymtabService.S_INSTANCES));

		while (insts.hasMoreElements()) {
		    ASTInstance inst = (ASTInstance)insts.nextElement();
		    res.add(inst);
		}
	    }
	}
	return(res.elements());
    }









    /**
     * Checks, if the data type used by this port is known.
     * If it has a primitive type, return true. Else, use the symtab to
     * lookup the data type.
     */
    private boolean checkPortType(ASTPort port) {
	boolean res = false;

	ASTDataTypeIdentifier dtid = port.getDataType();
	Assert.require(dtid);

	if (isUserDefinedDataType(dtid)) {
	    res = isDataTypeDeclared(dtid.getTypeID());
	} else {
	    res = isPrimitiveDataType(dtid);
	}

	return(res);
    }

    /**
     * Search the role with the given name in thegiven interface.
     */
    private ASTRole getRole(ASTInterface intf, ASTIdentifier id) {
	ASTRole res = null;

	if ((intf != null) && (id != null)) {

	    Enumeration e = checkEnum(SymtabService.getLocals(intf, SymtabService.S_ROLE));
	    while (e.hasMoreElements()) {
		ASTRole role = (ASTRole)e.nextElement();
		ASTIdentifier rid = role.getVarId();

		if (id.equals(rid)) {
		    res = role;
		    break;
		}
	    }
	}
	return(res);
    }

    /**
     * Search the given role definition in all of a components base interfaces.
     */
    private ASTRole getRoleDefinition(ASTComponent comp, ASTIdentifier id) {
	ASTRole res = null;

	if ((comp != null) && (id != null)) {

	    // and in its base interfaces
	    if (res == null) {
		Enumeration e = TypeAnalyzer.getBaseInterfaces(comp);

		while (e.hasMoreElements()) {
		    ASTIdentifier id2 = (ASTIdentifier)e.nextElement();
		    ASTInterface intf2 = SymtabService.lookupInterface(id2.getID());

		    if (intf2 != null) {
			res = getRole(intf2, id);
			if (res != null) {
			    break;
			}
		    }
		}
	    }
	}
	return(res);
    }
}





