package coco.semantic;


import java.util.*;

import coco.parser.ast.*;
import coco.parser.symtab.*;
import coco.utils.*;

/**
 * This class encapsulates the semantic tester/
 * It offers methods for testing various types of model elements
 */

public class SemanticService {


    // todo: component testing


    /**
     * This method checks all entities in the symbol table.
     */
    public static boolean testAllEntities() {
	boolean res = true;

	// start type analyzer
	prepare();

	res = res && testAllComponents();
	res = res && testAllInterfaces();
	res = res && testAllDataTypes();
	//	res = res && testAllRules();

	return(res);
    }

    /**
     * This method checks all components in the symbol table.
     */
    public static boolean testAllComponents() {
	boolean res = true;

	for (Enumeration enum = SymtabService.getComponents(); enum.hasMoreElements(); ) {
	    ASTComponent c = (ASTComponent)enum.nextElement();

	    if (!testComponent(c)) {
		res = false;
	    }
	}

	return(res);
    }

    /**
     * This method checks all interfaces in the symbol table.
     */
    public static boolean testAllInterfaces() {
	boolean res = true;

	for (Enumeration enum = SymtabService.getInterfaces(); enum.hasMoreElements(); ) {
	    ASTInterface c = (ASTInterface)enum.nextElement();

	    if (!testInterface(c)) {
		res = false;
	    }
	}

	return(res);
    }

    /**
     * This method checks all packets in the symbol table.
     */
    public static boolean testAllDataTypes() {
	boolean res = true;

	for (Enumeration enum = SymtabService.getDataTypes(); enum.hasMoreElements(); ) {
	    ASTDataType  c = (ASTDataType)enum.nextElement();

	    if (!testDataType(c)) {
		res = false;
	    }
	}

	return(res);
    }

    /**
     * This method checks all rules in the symbol table.
     */
    public static boolean testAllRules() {
	boolean res = true;

//  	for (Enumeration enum = SymtabService.getRules(); enum.hasMoreElements(); ) {
//  	    ASTRule c = (ASTRule)enum.nextElement();

//  	    if (!testRule(c)) {
//  		res = false;
//  	    }
//  	}

	return(res);
    }

    // single tests

    /**
     * This method runs tests on the supplied element.
     */
    public static boolean testNode(GenericNode n) {
	Assert.require(n);
	return(test(n));
    }

    /**
     * This method runs packet tests on the supplied element.
     */
    public static boolean testDataType(ASTDataType  p) {
	Assert.require(p);

	System.out.println("Check " + p.getDataTypeIdentifier().toString() + "...");
	//	System.err.println("Check " + p.getDataTypeIdentifier().toString() + "...");

	return(test(p));
    }

    /**
     * This method runs connector tests on the supplied element.
     */
    public static boolean testConnector(ASTConnector c) {
	Assert.require(c);
	return(test(c));
    }

    /**
     * This method runs rule tests on the supplied element.
     */
//      public static boolean testRule(ASTRule r) {
//  	Assert.require(r);
//  	return(test(r));
//      }

    /**
     * This method runs component/interface tests on the supplied element.
     */
    public static boolean testComponentType(ASTComponentType c) {
       	Assert.require(c);

	if (c instanceof ASTComponent) {

	    System.out.println("Check " + c.getID() + "...");
	    //	    System.err.println("Check " + c.getID() + "...");

	    return(testComponent((ASTComponent)c));
	} else if (c instanceof ASTInterface) {

	    System.out.println("Check " + c.getID() + "...");
	    //	    System.err.println("Check " + c.getID() + "...");

	    return(testInterface((ASTInterface)c));
	} else {
	    return(false);
	}
    }

    /**
     * This method runs interface tests on the supplied element.
     */
    public static boolean testInterface(ASTInterface c) {
	Assert.require(c);

	System.out.println("Check " + c.getID() + "...");
	//	System.err.println("Check " + c.getID() + "...");

	return(test(c));
    }

    /**
     * This method runs component tests on the supplied element.
     */
    public static boolean testComponent(ASTComponent c) {
	Assert.require(c);

	System.out.println("Check " + c.getID() + "...");
	//	System.err.println("Check " + c.getID() + "...");

	return(test(c));
    }

    /**
     * This is the real test method. It runs the test on the supplied node
     * and prints out the error messages.
     */
    private static boolean test(GenericNode n) {
	boolean res = false;

	Tester t = Tester.getTester(n);

	if ((t != null) && (t.isCompatible(n))) {
	    if (!(res = t.test(n))) {
		String[] errors = t.getErrors();
		System.out.println("check failed.");
		//		System.err.println("check failed.");
		System.err.println("begin error list...");

		for (int i = 0; i < errors.length; i++) {
		    System.err.println(errors[i]);
		}
		System.err.println("...end error list.");
		System.err.println();
	    } else {
		System.out.println("check succeeded.");
		//		System.err.println("check succeeded.");
	    }
	}

	return(res);
    }

    /**
     * A general preparation method, primary for adding all components and interfaces to
     * the type graph.
     */
    private static void prepare() {

	// reset analyzer

	TypeAnalyzer.reset();


	// add all components to the analyzer

	for (Enumeration enum = SymtabService.getComponents(); enum.hasMoreElements(); ) {
	    ASTComponent c = (ASTComponent)enum.nextElement();

	    TypeAnalyzer.addComponent(c);
	}

	// add all interfaces to the type analyzer

	for (Enumeration enum = SymtabService.getInterfaces(); enum.hasMoreElements(); ) {
	    ASTInterface c = (ASTInterface)enum.nextElement();

	    TypeAnalyzer.addInterface(c);
	}

	// add all data types

	for (Enumeration enum = SymtabService.getDataTypes(); enum.hasMoreElements(); ) {
	    ASTDataType  c = (ASTDataType)enum.nextElement();

	    TypeAnalyzer.addDataType(c);
	}

    }
}
